# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ErrorMessages.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='ErrorMessages.proto',
  package='xgt',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x13\x45rrorMessages.proto\x12\x03xgt\"J\n\x05\x45rror\x12 \n\x04\x63ode\x18\x01 \x01(\x0e\x32\x12.xgt.ErrorCodeEnum\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x0e\n\x06\x64\x65tail\x18\x03 \x01(\t*\xa2\x02\n\rErrorCodeEnum\x12\x0c\n\x08NO_ERROR\x10\x00\x12\x11\n\rGENERIC_ERROR\x10\x01\x12\x13\n\x0fNOT_IMPLEMENTED\x10\x02\x12\x12\n\x0eINTERNAL_ERROR\x10\x03\x12\x0c\n\x08IO_ERROR\x10\x04\x12\x17\n\x13SERVER_MEMORY_ERROR\x10\x05\x12\x14\n\x10\x43ONNECTION_ERROR\x10\x06\x12\x10\n\x0cSYNTAX_ERROR\x10\x07\x12\x0e\n\nTYPE_ERROR\x10\x08\x12\x0f\n\x0bVALUE_ERROR\x10\t\x12\x0e\n\nNAME_ERROR\x10\n\x12\x14\n\x10\x41RITHMETIC_ERROR\x10\x0b\x12\x1a\n\x16\x46RAME_DEPENDENCY_ERROR\x10\x0c\x12\x15\n\x11TRANSACTION_ERROR\x10\rb\x06proto3')
)

_ERRORCODEENUM = _descriptor.EnumDescriptor(
  name='ErrorCodeEnum',
  full_name='xgt.ErrorCodeEnum',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NO_ERROR', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GENERIC_ERROR', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NOT_IMPLEMENTED', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INTERNAL_ERROR', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IO_ERROR', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SERVER_MEMORY_ERROR', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CONNECTION_ERROR', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SYNTAX_ERROR', index=7, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TYPE_ERROR', index=8, number=8,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='VALUE_ERROR', index=9, number=9,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='NAME_ERROR', index=10, number=10,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ARITHMETIC_ERROR', index=11, number=11,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FRAME_DEPENDENCY_ERROR', index=12, number=12,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TRANSACTION_ERROR', index=13, number=13,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=105,
  serialized_end=395,
)
_sym_db.RegisterEnumDescriptor(_ERRORCODEENUM)

ErrorCodeEnum = enum_type_wrapper.EnumTypeWrapper(_ERRORCODEENUM)
NO_ERROR = 0
GENERIC_ERROR = 1
NOT_IMPLEMENTED = 2
INTERNAL_ERROR = 3
IO_ERROR = 4
SERVER_MEMORY_ERROR = 5
CONNECTION_ERROR = 6
SYNTAX_ERROR = 7
TYPE_ERROR = 8
VALUE_ERROR = 9
NAME_ERROR = 10
ARITHMETIC_ERROR = 11
FRAME_DEPENDENCY_ERROR = 12
TRANSACTION_ERROR = 13



_ERROR = _descriptor.Descriptor(
  name='Error',
  full_name='xgt.Error',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='xgt.Error.code', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message', full_name='xgt.Error.message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='detail', full_name='xgt.Error.detail', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28,
  serialized_end=102,
)

_ERROR.fields_by_name['code'].enum_type = _ERRORCODEENUM
DESCRIPTOR.message_types_by_name['Error'] = _ERROR
DESCRIPTOR.enum_types_by_name['ErrorCodeEnum'] = _ERRORCODEENUM
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Error = _reflection.GeneratedProtocolMessageType('Error', (_message.Message,), dict(
  DESCRIPTOR = _ERROR,
  __module__ = 'ErrorMessages_pb2'
  # @@protoc_insertion_point(class_scope:xgt.Error)
  ))
_sym_db.RegisterMessage(Error)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*- --------------------------------------------------===#
#
#  Copyright 2018-2019 Trovares Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and 
#  limitations under the License.
#
#===----------------------------------------------------------------------===#

import json
import os
import time
from datetime import timedelta, datetime
from os.path import expanduser

import grpc
import six

from . import ErrorMessages_pb2 as err_proto
from . import SchemaMessages_pb2 as sch_proto
from . import AdminService_pb2 as admin_proto
from . import AdminService_pb2_grpc as admin_grpc
from . import GraphTypesService_pb2 as graph_proto
from . import GraphTypesService_pb2_grpc as graph_grpc
from . import DataService_pb2 as data_proto
from . import DataService_pb2_grpc as data_grpc
from . import JobService_pb2 as job_proto
from . import JobService_pb2_grpc as job_grpc
from .common import (_assert_noerrors, _validated_property_name,
                     _validated_schema, _validated_frame_name,
                     _validate_opt_level, _to_unicode, Job,
                     XgtError, XgtNameError, XgtTypeError, XgtValueError,
                     XgtConnectionError, DEFAULT_OPT_LEVEL)
from .graph import TableFrame, VertexFrame, EdgeFrame
from .version import __version__

class Connection(object):
  """
  Connection to the server with functionality to create, change, and remove
  graph structures and run jobs.

  Parameters
  ----------
  host : str
    IP address of the computer where the server is running.
  port : int
    Port where the server is listening on for RPC calls.
  flags: dict
    Dictionary containing flags. Possible flags are:

    aws_access_key_id : str
      Amazon Access Key ID, used for authentication when loading data
      files from S3 buckets. The default is an empty string.
    aws_secret_access_key : str
      Amazon Access Key ID, used for authentication when loading data
      files from S3 buckets. The default is an empty string.
    ssl : boolean
      If true use ssl authentication for secure server channels.
      The default is False.
    ssl_root_dir : str
      Path to the root folder for ssl certificates and private keys.
      Defaults to the user's home directory.
    ssl_server_cn : str
      Common name on the certificate of the server to connect to.
      The default is the hostname.
  """

  def __init__(self, host='127.0.0.1', port=4367, flags=None):
    if flags is None:
      flags = {}
    self.port = port
    self.aws_access_key_id = flags.get('aws_access_key_id', '')
    self.aws_secret_access_key = flags.get('aws_secret_access_key', '')
    self.ssl = flags.get('ssl', False)
    self.ssl_root_dir = flags.get('ssl_root_dir', expanduser("~") + '/.ssl/')
    self.ssl_server_cn = flags.get('ssl_server_cn', host)
    self.host = host
    self.cwd = os.getcwd()
    self.client = six.moves.http_client
    self.current_opt_level = DEFAULT_OPT_LEVEL

    self._channel = self._create_channel()
    self._admin_svc = admin_grpc.AdminServiceStub(self._channel)
    self._graph_svc = graph_grpc.GraphTypesServiceStub(self._channel)
    self._data_svc = data_grpc.DataServiceStub(self._channel)
    self._job_svc = job_grpc.JobServiceStub(self._channel)

    self._server_unique_identifier = 0
    self._initialize_server_unique_identifier()
    self._version_check()

  def _create_channel(self):
    channel = None
    connection_string = self.host + ':' + _to_unicode(self.port)
    if (self.ssl):
      chain_cert = open(self.ssl_root_dir + '/certs/ca-chain.cert.pem', 'rb').read()
      client_key = open(self.ssl_root_dir + '/private/client.key.pem', 'rb').read()
      client_cert = open(self.ssl_root_dir + '/certs/client.cert.pem', 'rb').read()
      creds = grpc.ssl_channel_credentials(chain_cert, client_key, client_cert)
      try:
        channel = grpc.secure_channel(
            connection_string, creds,
            options=(('grpc.ssl_target_name_override', self.ssl_server_cn,),))
      except grpc.RpcError as ex:
        raise XgtConnectionError(ex.details(), '')
    else:
      try:
        channel = grpc.insecure_channel(connection_string)
      except grpc.RpcError as ex:
        raise XgtConnectionError(ex.details(), '')
    return channel

  def _initialize_server_unique_identifier(self):
    try:
      request = admin_proto.IdentifierRequest()
      response = self._run_rpc_function(request, self._admin_svc.Identifier)
      self._server_unique_identifier = response.server_unique_identifier
    except Exception:
      raise XgtConnectionError('Failure on connection initialization.')

  def call(self, request, rpc_function):
    response = self._run_rpc_function(request, rpc_function)
    try:
      _ = iter(response)
      # For response that return an iterator, it is the caller's responsibility
      # to check each packet for errors. E.g.:
      #   for result in response:
      #     _assert_noerrors(result)
      return response
    except TypeError:
      pass
    _assert_noerrors(response)
    return response

  def _run_rpc_function(self, request, rpc_function):
    try:
      _ = iter(request)
      # Iterators passed in as a request have already filled in the server id
      # per element of the iterator, so do not need to do it here.
    except TypeError:
      request.expected_server_id = self._server_unique_identifier
    try:
      response = rpc_function(request)
    except grpc.RpcError as ex:
      raise XgtConnectionError(ex.details(), '')
    return response

  def _convert_to_json(self, json_str):
    try:
      jsn = json.loads(json_str)
    except Exception as ex:
      raise XgtError("invalid json string: " + _to_unicode(ex) +
                     "\n==> " + _to_unicode(json_str[:1000]))
    if 'type' in jsn:
      if jsn['type'] == 'error' and 'data' in jsn and 'trace' in jsn and \
         'errorcode' in jsn:
        try:
          error_code_name = jsn['errorcode']
          error_class = _code_error_map[error_code_name]
          raise error_class(_to_unicode(jsn['data']),
                            _to_unicode(jsn['trace']))
        except XgtError:
          raise
        except Exception as ex:
          raise XgtError("Error detected while raising exception" +
                         _to_unicode(ex), _to_unicode(ex))
      elif jsn['type'] == 'error' and 'data' in jsn and 'trace' in jsn:
        raise XgtError(jsn['data'], jsn['trace'])
      elif jsn['type'] == 'error' and 'data' in jsn:
        raise XgtError(jsn['data'])
      elif jsn['type'] == "warning" and 'data' in jsn:
        print(jsn['data'])
    elif 'trace' in jsn:
      raise XgtError("Error: " + _to_unicode(jsn['error']),
                     _to_unicode(jsn['trace']))
    elif 'error' in jsn:
      raise XgtError("Error: " + _to_unicode(jsn['error']))
    return jsn

  def _version_check(self):
    server_version = None
    request = admin_proto.VersionRequest()
    response = self.call(request, self._admin_svc.Version)
    server_version = response.version
    if server_version != __version__:
      msg = "Version matching for xgt and " \
            "the server failed. Install the appropriate xgt " \
            "package. xGT version: " + server_version
      raise XgtError(msg)

  def _change_exit_error_count(self, action):
    action_u = action.upper()
    request = admin_proto.ChangeErrorCountRequest()
    request.action = admin_proto.ErrorCountActionEnum.Value(action_u)
    response = self.call(request, self._admin_svc.ChangeErrorCount)
    return response.error_count

  #------------------------- Housekeeping Methods
  @property
  def server_version(self):
    """
    Obtains the current product version from the server.

    Returns
    -------
    str
      Version number.

    """
    request = admin_proto.VersionRequest()
    response = self.call(request, self._admin_svc.Version)
    return response.version

  @property
  def max_user_memory_size(self):
    """
    Returns the maximum amount of memory available for user data on the xGT server.

    Returns
    -------
    int
      Maximum available user memory, in bytes.
    """
    request = admin_proto.MaxUserMemorySizeRequest()
    response = self.call(request, self._admin_svc.MaxUserMemorySize)
    return response.pool_size

  @property
  def free_user_memory_size(self):
    """
    Returns the amount of free memory available for user data on the xGT server.

    Returns
    -------
    int
      Currently available user memory, in bytes.
    """
    request = admin_proto.FreeUserMemorySizeRequest()
    response = self.call(request, self._admin_svc.FreeUserMemorySize)
    return response.free_memory_size

  def _stat_report(self):
    request = admin_proto.StatReportRequest()
    response = self.call(request, self._admin_svc.StatReport)
    return response.stat_report_json

  def _stat_reset(self):
    request = admin_proto.StatResetRequest()
    self.call(request, self._admin_svc.StatReset)

  #------------------------- Catalog Getter Methods
  def get_table_frames(self,names=None):
    """
    Get a list of TableFrame objects present in the server.

    A TableFrame object allows for interaction with a table present in the xGT server.
    For example, a table may be created by a MATCH query and may contain query results.
    It may also be explicitly created with `Connection.create_table_frame()`.

    Parameters
    ----------
    names : list of strings or None
      If a list, the list of names of tables frames to retrieve.
      If None, all table frames are returned.

    Returns
    -------
    list
      TableFrame objects representing tables present in the server.

    Examples
    --------
    >>> conn = xgt.Connection()
    >>> ... create graph
    >>> qr1 = 'MATCH (a:Employee) RETURN a.PersonID INTO Result1'
    >>> conn.run_job(qr1)
    >>> table_frames = conn.get_table_frames()
    >>> print [f.name for f in table_frames]
    ['Employee_table', 'ReportsTo_table', 'Result1']
    >>> result1_data = table_frames[2].get_data_pandas()

    """
    if names is None:
      names = []
    elif isinstance(names, six.string_types):
      raise TypeError('Invalid argument: "names" must be a list of strings')
    else:
      names = [_validated_frame_name(n) for n in names]

    request = graph_proto.GetTableFramesRequest()
    if names is not None:
      request.name.extend(names)
    response = self.call(request, self._graph_svc.GetTableFrames)
    frames = []
    for data in response.table_type:
      schema = []
      for prop in data.schema.property:
        prop_type = sch_proto.UvalTypeEnum.Name(prop.data_type).lower()
        prop_type = _to_unicode(prop_type)
        schema.append([prop.name, prop_type])
      frames.append(TableFrame(self, data.name, schema, data.error_name))
    return frames

  def get_table_frame(self, name):
    """
    Get a TableFrame object that allows interaction with a table present in the xGT server.

    A TableFrame object allows for interaction with a table present in the xGT server.
    For example, a table may be created by a MATCH query and may contain query results.
    It may also be explicitly created with `Connection.create_table_frame()`.

    Parameters
    ----------
    name : str
      Table name.

    Returns
    -------
    TableFrame
      Frame to the table.

    Examples
    --------
    >>> conn = xgt.Connection()
    >>> ... create graph and run queries ...
    >>> t = conn.get_table_frame('Employee_table')
    >>> print(str(t))
    {
      'name': 'Employee_table',
      'schema': [
        ['PersonID', 'int'],
        ['Name', 'text'],
        ['PostalCode', 'int']]
    }
    >>> qr1 = 'MATCH (a:Employee) RETURN a.PersonID INTO Result1'
    >>> conn.run_job(qr1)
    >>> result1_table = conn.get_table_frame('Result1')
    >>> num_results = result1_table.num_rows
    >>> result1_data = result1_table.get_data_pandas()

    """
    frames = self.get_table_frames([name])
    return frames[0]

  def get_vertex_frames(self, names = None):
    """
    Get a list of VertexFrame objects present in the xGT server.

    A VertexFrame represents a collection of vertices held on the xGT
    server and can be used to retrieve information about them.
    `VertexFrame.get_data_pandas()` and `VertexFrame.get_data()` are
    used to retrieve member vertices.
    Each vertex in a VertexFrame shares the same properties,
    described in `VertexFrame.schema`. Each vertex in a VertexFrame
    is uniquely identified by the property listed in `VertexFrame.key`.

    Parameters
    ----------
    names : list of strings or None
      If a list, the list of names of vertex frames to retrieve.
      If None, all vertex frames are returned.

    Returns
    -------
    list
      VertexFrame objects present in the server.

    Examples
    --------
    >>> conn = xgt.Connection()
    >>> print [f.name for f in conn.get_vertex_frames()]
    ['Company', 'Person']
    >>> print [f.num_vertices for f in conn.get_vertex_frames()]
    [3, 101]

    """
    if names is None:
      names = []
    elif isinstance(names, six.string_types):
      raise TypeError('Invalid argument: "names" must be a list of strings')
    else:
      names = [_validated_frame_name(n) for n in names]

    request = graph_proto.GetVertexFramesRequest()
    request.name.extend(names)
    response = self.call(request, self._graph_svc.GetVertexFrames)
    frames = []
    for data in response.vertex_type:
      schema = []
      for prop in data.schema.property:
        prop_type = sch_proto.UvalTypeEnum.Name(prop.data_type).lower()
        prop_type = _to_unicode(prop_type)
        schema.append([prop.name, prop_type])
      vertex_key_val = sch_proto.RoleEnum.Value('VERTEX_KEY')
      key = None
      for prop in data.schema.property:
        prop_type = sch_proto.UvalTypeEnum.Name(prop.data_type).lower()
        prop_type = _to_unicode(prop_type)
        if prop.role == vertex_key_val:
          key = prop.name
          break
      frames.append(VertexFrame(self, data.name, schema, key, data.error_name))
    return frames

  def get_vertex_frame(self, name):
    """
    Get a VertexFrame object that allows interaction with a collection of vertices.

    A VertexFrame represents a collection of vertices held on the xGT
    server and can be used to retrieve information about them.
    `VertexFrame.get_data_pandas()` and `VertexFrame.get_data()` are
    used to retrieve member vertices.
    Each vertex in a VertexFrame shares the same properties,
    described in `VertexFrame.schema`. Each vertex in a VertexFrame
    is uniquely identified by the property listed in `VertexFrame.key`.

    Parameters
    ----------
    name : str
      VertexFrame name.

    Returns
    -------
    VertexFrame
      Frame to the collection of vertices.

    Examples
    --------
    >>> conn = xgt.Connection()
    >>> v = conn.get_vertex_frame('Person')
    >>> print(str(v))
    {
      'name': 'Person',
      'key': 'id',
      'schema': [
        ['id', 'int'],
        ['name', 'text']],
    }
    >>> print(str(v.num_vertices))
    101
    >>> vertices = v.get_data_pandas()

    """
    frames = self.get_vertex_frames([name])
    return frames[0]

  def get_edge_frames(self, names=None, graph=None):
    """
    Get a list of EdgeFrame objects present in the xGT server.

    An EdgeFrame represents a collection of edges held on the xGT
    server and can be used to retrieve information about them.
    `EdgeFrame.get_data_pandas()` and `EdgeFrame.get_data()` are
    used to retrieve member edges.
    Edge edge in an EdgeFrame shares the same properties, described
    in `EdgeFrame.schema`.

    Parameters
    ----------
    names : list of strings or None
      If a list, the list of names of edge frames to retrieve.
      If None, all edge frames are returned.

    Returns
    -------
    list
      EdgeFrame objects present in the server.

    Examples
    --------
    >>> conn = xgt.Connection()
    >>> print [f.name for f in conn.get_edge_frames()]
    ['RelatedTo', 'WorksFor']

    """
    if names is None:
      names = []
    elif isinstance(names, six.string_types):
      raise TypeError('Invalid argument: "names" must be a list of strings')
    else:
      names = [_validated_frame_name(n) for n in names]

    request = graph_proto.GetEdgeFramesRequest()
    request.name.extend(names)
    response = self.call(request, self._graph_svc.GetEdgeFrames)
    frames = []
    for data in response.edge_type:
      schema = []
      for prop in data.schema.property:
        prop_type = sch_proto.UvalTypeEnum.Name(prop.data_type).lower()
        prop_type = _to_unicode(prop_type)
        schema.append([prop.name, prop_type])
      source_key = data.source_key
      target_key = data.target_key
      frames.append(EdgeFrame(self, data.name, schema,
                              data.source_vertex, data.target_vertex,
                              source_key, target_key, data.error_name))
    return frames

  def get_edge_frame(self, name):
    """
    Get an EdgeFrame object that allows interaction with a collection of edges.

    An EdgeFrame represents a collection of edges held on the xGT
    server and can be used to retrieve information about them.
    `EdgeFrame.get_data_pandas()` and `EdgeFrame.get_data()` are
    used to retrieve member edges.
    Edge edge in an EdgeFrame shares the same properties, described
    in `EdgeFrame.schema`.

    Parameters
    ----------
    name : str
      EdgeFrame name.

    Returns
    -------
    EdgeFrame
      Frame to the collection of edges.

    Examples
    --------
    >>> conn = xgt.Connection()
    >>> ... create graph and run queries ...
    >>> e = conn.get_edge_frame('WorksFor')
    >>> print(str(e))
    {
      'name': 'WorksFor',
      'source': 'Person',
      'target': 'Company',
      'schema': [
        ['srcid', 'int'],
        ['trgid', 'int']],
      'source_key' : 'srcid',
      'target_key' : 'trgid'
    }
    >>> edges = e.get_data_pandas()

    """
    frames = self.get_edge_frames([name])
    return frames[0]

  #------------------------- DDL Methods
  def create_table_frame(self, name, schema):
    """
    Create a new TableFrame in the server.

    A TableFrame object represents a table held on the xGT server and can be
    used to retrieve information about it. The TableFrame schema describes
    the names and data types of table properties.

    Parameters
    ----------
    name : str
      Name of table.
    schema : list of pairs
      List of pairs associating property names with xGT data types.

    Returns
    -------
    TableFrame
      Frame to the table.

    Examples
    --------
    >>> import xgt
    >>> conn = xgt.Connection()
    >>> conn.create_table_frame(
          name = 'tab1',
    ...   schema = [['id', xgt.INT],
    ...             ['name', xgt.TEXT]])

    """
    name = _validated_frame_name(name)
    schema = _validated_schema(schema)

    request = graph_proto.CreateTableFrameRequest()
    request.table_type.name = name
    for col_name,col_type in schema:
      prop = sch_proto.Property()
      prop.name = col_name
      prop.data_type = sch_proto.UvalTypeEnum.Value(col_type)
      prop.role = sch_proto.RoleEnum.Value('PROPERTY')
      request.table_type.schema.property.extend([prop])
    response = self.call(request, self._graph_svc.CreateTableFrame)
    data = response.table_type[0]
    schema = []
    for prop in data.schema.property:
      prop_type = sch_proto.UvalTypeEnum.Name(prop.data_type).lower()
      prop_type = _to_unicode(prop_type)
      schema.append([prop.name, prop_type])
    frame = TableFrame(self, data.name, schema, data.error_name)
    return frame

  def create_vertex_frame(self, name, schema, key):
    """
    Create a new VertexFrame in the server.

    A VertexFrame represents a grouping or collection of vertices
    held on the xGT server, all sharing the same property names
    and types. This function creates a new frame of vertices
    on the xGT server and returns a VertexFrame representing it.

    Parameters
    ----------
    name : str
      Name of vertex frame.
    schema : list of pairs
      List of pairs associating property names with xGT data types.
      Each vertex in the VertexFrame will have these properties.
    key : str
      The property name used to uniquely identify vertices in the graph.
      This is the name of one of the properties from the schema and
      must be unique for each vertex in the frame.

    Returns
    -------
    VertexFrame
      Frame to the collection of vertices.

    Examples
    --------
    >>> import xgt
    >>> conn = xgt.Connection()
    >>> pers_vtx = conn.create_vertex_frame(
                     name = 'Person',
    ...              schema = [['id', xgt.INT],
    ...                        ['name', xgt.TEXT]],
    ...              key = 'id')

    """
    name = _validated_frame_name(name)
    schema = _validated_schema(schema)
    key = _validated_property_name(key)

    key_found = False
    if key not in [prop for prop,_ in schema]:
      msg = u'The vertex key "{0}" does not match any schema property ' \
            u'name in this frame.'
      raise XgtNameError(msg.format(key))

    request = graph_proto.CreateVertexFrameRequest()
    request.vertex_type.name = name
    for col_name,col_type in schema:
      prop = sch_proto.Property()
      prop.name = col_name
      prop.data_type = sch_proto.UvalTypeEnum.Value(col_type)
      if col_name == key:
        prop.role = sch_proto.RoleEnum.Value('VERTEX_KEY')
      else:
        prop.role = sch_proto.RoleEnum.Value('PROPERTY')
      request.vertex_type.schema.property.extend([prop])
    response = self.call(request, self._graph_svc.CreateVertexFrame)
    frame = None
    data = response.vertex_type[0]
    schema = []
    for prop in data.schema.property:
      prop_type = sch_proto.UvalTypeEnum.Name(prop.data_type).lower()
      prop_type = _to_unicode(prop_type)
      schema.append([prop.name, prop_type])
    vertex_key_val = sch_proto.RoleEnum.Value('VERTEX_KEY')
    key = None
    for prop in data.schema.property:
      prop_type = sch_proto.UvalTypeEnum.Name(prop.data_type).lower()
      prop_type = _to_unicode(prop_type)
      if prop.role == vertex_key_val:
        key = prop.name
        break
    frame = VertexFrame(self, data.name, schema, key, data.error_name)
    return frame

  def create_edge_frame(self, name, schema, source, target, source_key, target_key):
    """
    Create a new EdgeFrame in the server.

    An EdgeFrame represents a collection of edges held on the xGT server
    that share the same property names and types. The source vertex
    of each edge in an EdgeFrame must belong to the same VertexFrame.
    This source VertexFrame is identified by the source parameter of
    this function. The target vertex of each edge in an EdgeFrame must
    belong to the same VertexFrame. This target VertexFrame is identified
    by the target parameter.

    For each edge in the EdgeFrame, its source vertex is identified by
    the edge property name given in the parameter source_key, which must
    be one of the properties listed in the schema. The edge target vertex
    is identified by the property name given in the parameter target_key,
    which must be one of the properties listed in the schema.

    Parameters
    ----------
    name : str
      Name of edge frame.
    schema : list of pairs
      List of pairs associating property names with xGT data types.
      Each edge in the EdgeFrame will have these properties.
    source : str or VertexFrame
      The name of a VertexFrame or a VertexFrame object.
      The source vertex of each edge in this EdgeFrame will belong
      to this VertexFrame.
    target : str or VertexFrame
      The name of a VertexFrame or a VertexFrame object.
      The target vertex of each edge in this EdgeFrame will belong
      to this VertexFrame.
    source_key : str
      The edge property name that identifies the source vertex of an edge.
      This is one of the properties from the schema.
    target_key : str
      The edge property name that identifies the target vertex of an edge.
      This is one of the properties from the schema.

    Returns
    -------
    EdgeFrame
      Frame to the collection of edges.

    Examples
    --------
    >>> import xgt
    >>> conn = xgt.Connection()
    >>> conn.create_vertex_frame(
          name = 'Person',
    ...   schema = [['id', xgt.INT],
    ...             ['name', xgt.TEXT]],
    ...   key = 'id')
    >>> conn.create_vertex_frame(
          name = 'Company',
    ...   schema = [['id', xgt.INT],
    ...             ['size', xgt.TEXT],
    ...             ['name', xgt.TEXT]],
    ...   key = 'id')
    >>> conn.create_edge_frame(
          name = 'WorksFor',
    ...   schema = [['srcid', xgt.INT],
    ...             ['role', xgt.TEXT],
    ...             ['trgid', xgt.INT]],
    ...   source = 'Person',
    ...   target = 'Company',
    ...   source_key = 'srcid',
    ...   target_key = 'trgid')

    """
    name = _validated_frame_name(name)
    schema = _validated_schema(schema)
    source_key = _validated_property_name(source_key)
    target_key = _validated_property_name(target_key)

    if source_key not in [prop for prop, _ in schema]:
      msg = u'The source key "{0}" does not match any schema property ' \
            u'name in this frame.'
      raise XgtNameError(msg.format(source_key))
    if target_key not in [prop for prop, _ in schema]:
      msg = u'The target key "{0}" does not match any schema property ' \
            u'name in this frame.'
      raise XgtNameError(msg.format(target_key))

    if isinstance(source, VertexFrame):
      source_name = source.name
    else:
      source_name = _validated_frame_name(source)
    if isinstance(target, VertexFrame):
      target_name = target.name
    else:
      target_name = _validated_frame_name(target)

    request = graph_proto.CreateEdgeFrameRequest()
    request.edge_type.name = name
    request.edge_type.source_vertex = source_name
    request.edge_type.target_vertex = target_name
    request.edge_type.source_key = source_key
    request.edge_type.target_key = target_key
    for col_name, col_type in schema:
      prop = sch_proto.Property()
      prop.name = col_name
      prop.data_type = sch_proto.UvalTypeEnum.Value(col_type)
      if col_name == source_key:
        prop.role = sch_proto.RoleEnum.Value('EDGE_SOURCE_KEY')
      elif col_name == target_key:
        prop.role = sch_proto.RoleEnum.Value('EDGE_TARGET_KEY')
      else:
        prop.role = sch_proto.RoleEnum.Value('PROPERTY')
      request.edge_type.schema.property.extend([prop])

    response = self.call(request, self._graph_svc.CreateEdgeFrame)
    data = response.edge_type[0]
    schema = []
    for prop in data.schema.property:
      prop_type = sch_proto.UvalTypeEnum.Name(prop.data_type).lower()
      prop_type = _to_unicode(prop_type)
      schema.append([prop.name, prop_type])
    frame = EdgeFrame(self, data.name, schema,
                      data.source_vertex, data.target_vertex, data.source_key,
                      data.target_key, data.error_name)
    return frame

  def drop_frame(self, frame):
    """
    Drop a VertexFrame, EdgeFrame, or TableFrame.

    Parameters
    ----------
    frame : str, VertexFrame, EdgeFrame, or TableFrame
      A frame or the name of frame to drop on the xGT server.

    Returns
    -------
    bool
      True if frame was found and dropped and False if frame was not found.

    """
    if isinstance(frame, TableFrame):
      name = frame.name
    else:
      name = _validated_frame_name(frame)
    request = graph_proto.DeleteTypeRequest()
    request.name = name
    response = self.call(request, self._graph_svc.DeleteType)
    return response.found_and_deleted

  #------------------------- Job Methods

  def set_optimization_level(self, optlevel = DEFAULT_OPT_LEVEL):
    """
    Set the optimization level for TQL queries.

    Parameters
    ----------
    optlevel : int
      The optimization level values are:
        - 0: No optimization.
        - 1: General optimization.
        - 2: WHERE-clause optimization.

      Optional. Default is 2'.
    """
    if _validate_opt_level(optlevel):
      self.current_opt_level = optlevel

  def get_jobs(self, jobids=None):
    """
    Get a list of Job objects, each representing the state of
    the job on the server at the point in time of the
    invocation of this function.

    Parameters
    ----------
    jobids : list of ints
      A list of job ids for which to return Job objects.
      By default all jobs are returned.

    Returns
    -------
    list
      A list of Job objects, each representing the state
      of a job in the server.

    Examples
    --------
    >>> conn = xgt.Connection()
    >>> ... create vertices and edges and run queries ...
    >>> all_jobs = conn.get_jobs()
    >>> for j in all_jobs:
    >>> ... print j
    id:6, status:completed
    id:7, status:completed
    id:8, status:running

    """

    if jobids is None:
      request = job_proto.GetJobsRequest()
      response = self.call(request, self._job_svc.GetJobs)
      jobids = []
      for s in response.job_status:
        jobids.append(s.job_id)
    return [Job(self, i) for i in jobids]

  def cancel_job(self, job):
    """
    Cancel the execution of a job in the server.

    A job can be canceled only if it is *running* and will have a status of
    *canceled* after its cancellation. A job that already had a status of
    *completed* or *failed* before invoking this function will keep that
    status after invoking this function.

    Parameters
    ----------
    job : Job, int
      A Job object or an integer job id to cancel.

    Returns
    -------
    bool
      True if the job was cancelled. False if the job already had a
      status of completed or failed before invoking this function.

    Examples
    --------
    >>> conn = xgt.Connection()
    >>> ... create vertices and edges and run queries ...
    >>> print(conn.cancel_job(18))
    id:18, status:completed
    >>> all_jobs = conn.get_jobs()
    >>> for j in all_jobs:
    >>> ... conn.cancel_job(j)

    """
    if isinstance(job, Job):
      jobid = job.id
    elif isinstance(job, six.integer_types):
      jobid = job
    else:
      raise TypeError("Job must be a Job object or an int.")

    # Get job status.
    request = job_proto.GetJobsRequest()
    request.job_id.extend([jobid])
    response = self.call(request, self._job_svc.GetJobs)
    # Cancel job if it's not in a terminal state.
    if len(response.job_status) > 0:
      job_status = response.job_status[0]
      if job_status.status in [job_proto.SCHEDULED, job_proto.RUNNING]:
        request = job_proto.CancelJobsRequest()
        request.job_id.extend([jobid])
        self.call(request, self._job_svc.CancelJobs)
        return True
      else:
        return False
    else:
      return False

  def run_job(self, query, optlevel=None, timeout=None):
    """
    Run a TQL query as a job. This function blocks
    until the job stops running.

    Parameters
    ----------
    queries : str
      One TQL query string.
    optlevel : int
      The optimization level values are:
        - 0: No optimization.
        - 1: General optimization.
        - 2: WHERE-clause optimization.

      Optional. Default=None, which implies a value of '2'.
    timeout : int
      Maximum number of seconds that the query should take before being
      automatically canceled.
      Optional. Default=None where an infinite value is assumed.

    Returns
    -------
    Job
      A Job object for the query.

    Raises
    -------
    XgtError
      If the query is not a string (str) or the query text is
      larger than 209,000 characters.

    Examples
    --------
    >>> conn = xgt.Connection()
    >>> ... create vertices and edges ...
    >>> job = conn.run_job('MATCH (a:Employee) RETURN a.PersonID INTO Result1')
    >>> print(job)
    id:20, status:completed

    >>> conn.run_job('MATCH (a) RETURN a.id INTO Result1')
    ...
    xgt.common.XgtError: Failed jobs:
    id: 22
    error:
    VertexFrameManager: No object registered with this ObjectID 18446744073709551615

    """
    job = self.schedule_job(query, optlevel)
    return  self.wait_for_job(job, timeout)

  def schedule_job(self, query, optlevel=None):
    """
    Schedule a TQL query as a job. This function
    returns immediately after scheduling the job.

    Parameters
    ----------
    query : str
      One TQL query string.
    optlevel : int
      The optimization level values are:
        - 0: No optimization.
        - 1: General optimization.
        - 2: WHERE-clause optimization.


    Returns
    -------
    Job
      A Job object representing the job that has been scheduled.

    Raises
    -------
    XgtError
      If the query is not a string (str) or the query text is
      larger than 209,000 characters.

    Examples
    --------
    >>> conn = xgt.Connection()
    >>> ... create vertices and edges ...
    >>> query = 'MATCH (a:Employee) RETURN a.PersonID INTO Result1'
    >>> job = conn.schedule_job(query)
    >>> print(job)
    id:25, status:scheduled

    """
    if not isinstance(query, six.string_types):
      raise TypeError(
          "Unexpected argument type '" + _to_unicode(type(query)) + "'")
    if optlevel is None:
      optlevel = self.current_opt_level
    elif _validate_opt_level(optlevel):
      pass

    request = job_proto.ScheduleJobsCypherRequest()
    request.cypher_query.extend([query])
    request.optimization_level = optlevel
    response = self.call(request, self._job_svc.ScheduleJobsCypher)
    one_job = response.job_status[0]
    return Job(self, one_job.job_id)

  def wait_for_job(self, job, timeout=None):
    """
    Wait for a job. This function blocks until the job stops running.

    Parameters
    ----------
    job : Job, int
      A Job object or an integer job id.
    timeout : int
      Number of seconds each job is allowed to execute before being
      automatically cancelled.
      Optional. Default=None (no timeout).

    Returns
    -------
    Job
      A Job object representing the state of the job on the server.

    Raises
    -------
    XgtError
      If the query is not a string (str) or the query text is
      larger than 209,000 characters.
      If one or more query jobs failed.

    Examples
    --------
    >>> conn = xgt.Connection()
    >>> ... create vertices and edges ...
    >>> qr1 = 'MATCH (a:Employee) RETURN a.PersonID INTO Result1'
    >>> jb1 = conn.schedule_job(qr1)
    >>> qr2 = 'MATCH (b:Employee) RETURN b.PersonID INTO Result2'
    >>> jb2 = conn.schedule_job(qr2)
    >>> jb1 = conn.wait_for_job(jb1)
    >>> print(jb1)
    id:31, status:completed
    >>> jb2 = conn.wait_for_job(jb2)
    >>> print(jb2)
    id:32, status:completed

    """
    if isinstance(job, Job):
      jobid = job.id
    elif isinstance(job, six.integer_types):
      jobid = job
    else:
      raise TypeError("Job must be a Job object or an int.")
    begin = datetime.utcnow()
    endtime = self._calculate_endtime(begin, timeout)
    check_interval_sec = 0.1
    loop = True
    while True:
      status = ''
      request = job_proto.GetJobsRequest()
      request.job_id.extend([jobid])
      response = self.call(request, self._job_svc.GetJobs)
      job = response.job_status[0]  # Retrieve only one job from the list.
      status = job_proto.JobStatusEnum.Name(job.status).lower()
      loop = loop and (status in ['scheduled', 'started', 'running',
                                  'transactionnotstarted',
                                  'transactioncommitting'])
      if not endtime is None:
        end = datetime.utcnow()
        if end >= endtime:
          was_cancelled = self.cancel_job(jobid)
      if not loop:
        break
      time.sleep(check_interval_sec)
      if check_interval_sec < 10.0:
        check_interval_sec = check_interval_sec + 0.1
    job_obj = Job(self, jobid)
    if job_obj.status == 'failed':
      msg = (u'Failed job. id={0} msg="{1}"').format(jobid, job_obj.error)
      raise job_obj.error_type(msg, job_obj.trace)
    return job_obj

  def _calculate_endtime(self, begin, timeout):
    endtime = None
    if not timeout is None:
      try:
        endtime = begin + timedelta(seconds=timeout)
      except TypeError:
        msg = ('Expected an int value for “timeout” argument, ' +
               'not "{0}"').format(_to_unicode(timeout))
        raise TypeError(msg)
    return endtime



# -*- coding: utf-8 -*- --------------------------------------------------===#
#
#  Copyright 2018-2019 Trovares Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and 
#  limitations under the License.
#
#===----------------------------------------------------------------------===#

import collections
import csv
import json
import math
import sys
import urllib

import grpc
import six

from . import ErrorMessages_pb2 as err_proto
from . import SchemaMessages_pb2 as sch_proto
from . import GraphTypesService_pb2 as graph_proto
from . import GraphTypesService_pb2_grpc as graph_grpc
from . import DataService_pb2 as data_proto
from . import DataService_pb2_grpc as data_grpc
from .common import (_assert_noerrors, _to_bytes, _to_str, _to_unicode,
                     _unicode_to_str, XgtNotImplemented, XgtIOError,
                     MAX_PACKET_SIZE)

class HeaderMode:
  NONE = 'none'
  IGNORE = 'ignore'
  NORMAL = 'normal'
  STRICT = 'strict'

  _all = [NONE,IGNORE,NORMAL,STRICT]

# -----------------------------------------------------------------------------

class TableFrame(object):
  """
  TableFrame object represent a table held on the xGT server; it can be
  used to retrieve information about it and should not be instantiated directly
  by the user. Methods that return this object: `Connection.get_table_frame()`,
  `Connection.get_table_frames()` and `Connection.create_table_frame()`. A table
  may also be created by a MATCH query and may contain query results.

  Parameters
  ----------
  conn : Connection
    An open connection to an xGT server.
  name : str
    Name of table.
  schema : list of pairs
    List of pairs associating property names with xGT data types.
  error_frame_name : str
    Name of the frame created to store errors encountered while loading data.
    (internal)

  Examples
  --------
  >>> import xgt
  >>> conn = xgt.Connection()
  >>> ... run query and store results in table01
  >>> t = conn.get_table_frame('table01')
  >>> print(t.name)

  """
  def __init__(self, conn, name, schema, error_frame_name):
    self._conn = conn
    self._name = name
    self._schema = [[c[0], c[1]] for c in schema]
    self._error_frame_name = error_frame_name

  @property
  def name(self):
    """str: Name of the table."""
    return self._name

  def error_frame_name(self):
    """str: Name of the error table this frame will produce on errors."""
    return self._error_frame_name

  @property
  def schema(self):
    """list of lists: Gets the property names and types of the table."""
    return self._schema

  def load(self, paths, headerMode=HeaderMode.NONE):
    """
    Loads data from one or more CSV files specified in the list of paths.
    Each path may have its own protocol as described below.

    Parameters
    ----------
    paths : list
      Paths to the CSV files.

      ==================== =====================================
                      Syntax for one CSV file path
      ----------------------------------------------------------
          Resource type                 Path syntax
      ==================== =====================================
          local to python: '<path to csv file>'
                           'xgt://<path to csv file>'
          xgtd computer:   'xgtd://<path to csv file>'
          AWS s3:          's3://<path to csv file>'
          https site:      'https://<path to csv file>'
          http site:       'http://<path to csv file>'
          ftps server:     'ftps://<path to csv file>'
          ftp server:      'ftp://<path to csv file>'
      ==================== =====================================

    headerMode : str
      Indicates if the files contain headers:
        - HeaderMode.NONE
        - HeaderMode.IGNORE
        - HeaderMode.NORMAL
        - HeaderMode.STRICT

      Optional. Default=HeaderMode.NONE.

    """
    if paths is None:
      raise TypeError('the "paths" parameter is None')
    if headerMode is None:
      raise TypeError('the "headerMode" parameter is None')
    if headerMode not in HeaderMode._all:
      raise TypeError('Invalid header mode: "{0}"'.format(_to_str(headerMode)))
    if not isinstance(paths, (list, tuple, six.string_types)):
      raise TypeError('one or more file paths are expected; the data type of the "paths" parameter is "{0}"'.format(type(paths)))
    client_paths, server_paths, url_paths = _group_paths(paths, True)
    if len(client_paths) == 0 and len(server_paths) == 0 and len(url_paths) == 0:
      raise XgtIOError('no valid paths found: ' + _to_unicode(paths))
    if len(client_paths) > 0:
      self._insert_from_csv(client_paths, headerMode)
    if len(server_paths) > 0:
      self._ingest(server_paths, headerMode)
    if len(url_paths) > 0:
      self._ingest(url_paths, headerMode)

  def save(self, path, offset=0, length=None, headers=False):
    """
    Writes the rows from the frame to a CSV file in the path and the
    computer indicated by the path.

    Parameters
    ----------
    path : str
      Path to the CSV file.

      ==================== =====================================
                      Syntax for one CSV file path
      ----------------------------------------------------------
          Resource type                 Path syntax
      ==================== =====================================
          local to python: '<path to csv file>'
                           'xgt://<path to csv file>'
          xgtd computer:   'xgtd://<path to csv file>'
      ==================== =====================================

    offset : int
      Position (index) of the first row to be retrieved.
      Optional. Default=0.
    length : int
      Maximum number of rows to be retrieved.
      Optional. Default=None.
    headers : boolean
      Indicates if headers should be added.
      Optional. Default=False.

    """
    if path is None:
      raise TypeError('the "paths" parameter is None')
    if not isinstance(path, six.string_types):
      msg = 'a file path is expected; the data type of the "path" ' \
            'parameter is "{0}"'.format(type(path))
      raise TypeError(msg)

    client_paths, server_paths, url_paths = _group_paths(path, False)
    if (len(client_paths) == 0 and len(server_paths) == 0 and
        len(url_paths) == 0):
      raise XgtIOError('no valid paths found: ' + _to_unicode(path))
    if len(client_paths) > 0:
      self._save_to_csv(client_paths[0], offset, length, headers)
    if len(server_paths) > 0:
      self._egest(server_paths[0], headers)

  def _insert_packet_generator(self, data, is_pandas):
    columns = []
    column_types = {}
    for item in self.schema:
      columns.append(item[0])
      column_types[item[0]] = item[1]
    schemasize = len(columns)
    nrow = len(data)
    col_names = '"' + ('","'.join(columns)) + '"'

    tbl = ''
    for i in range(nrow):
      row = ''
      data_row = ''
      if is_pandas:
        length = len(data.columns)
      else:
        length = len(data[i])
      for j in range(length):
        if is_pandas:
          col = data[columns[j]][i]
        else:
          col = data[i][j]
        if j < schemasize:
          col_type = column_types[columns[j]]
          if col is None:
            strcol = ''
          elif isinstance(col, float) and math.isnan(col):
            strcol = ''
          else:
            strcol = _to_unicode(col)
            if isinstance(col, bool):
                strcol = strcol.lower()
            if isinstance(col, six.string_types):
                strcol = '"' + strcol + '"'
          row = row + strcol + ','
        else:
          row = row + _to_unicode(col) + ','
      tbl = tbl + row[:-1] + '\n'
      if len(tbl) >= MAX_PACKET_SIZE:
        request = data_proto.UploadDataRequest()
        request.repository_name = self._name
        request.content_type = data_proto.CSV
        request.header_mode = data_proto.NONE
        request.content = tbl.encode('utf-8')
        request.expected_server_id = self._conn._server_unique_identifier
        yield request
        tbl = ''
    if len(tbl) > 0:
      request = data_proto.UploadDataRequest()
      request.repository_name = self._name
      request.content_type = data_proto.CSV
      request.header_mode = data_proto.NONE
      request.content = tbl.encode('utf-8')
      request.expected_server_id = self._conn._server_unique_identifier
      yield request
      tbl = ''

  def insert(self, data):
    """
    Inserts data rows. The properties of the new data must match the schema
    in both order and type.

    Parameters
    ----------
    data : list or Pandas dataframe
      Data represented by a list of lists of data items or by a
      Pandas Dataframe.

    """
    if data is None:
      return
    if len(data) == 0:
      return

    #---- if user passed in a pandas frame
    is_pandas = False
    try:
      import pandas
      if isinstance(data, pandas.DataFrame):
        is_pandas = True
    except:
      pass

    # Exceptions for iterators get eaten by grpc so we check outside
    # the generator function:
    if not is_pandas:
      if (isinstance(data, six.text_type) or
          not isinstance(data, collections.Iterable)):
        raise TypeError('a list of lists or a Pandas DataFrame is expected')
      for i in range(len(data)):
          if not isinstance(data[i], collections.Iterable):
              msg = 'Row #{0} is not a list. A list of lists ' \
                    'or a Pandas DataFrame is expected'.format(i)
              raise TypeError(msg)

    data_iter = self._insert_packet_generator(data, is_pandas)

    self._conn.call(data_iter, self._conn._data_svc.UploadData)

  def get_data(self, offset=0, length=None):
    """
    Returns frame data starting at a given offset and spanning a given
    length.

    Parameters
    ----------
    offset : int
      Position (index) of the first row to be retrieved.
      Optional. Default=0.
    length : int
      Maximum number of rows to be retrieved starting from the row
      indicated by offset. A value of 'None' means 'all rows' on and
      after the offset.
      Optional. Default=None.

    Returns
    -------
    list of lists

    """
    # Uses json conversion which is slow, res can get huge
    # if the frame data is huge.
    responses = self._get_data(offset=offset, length=length)
    res = ''
    for response in responses:
        _assert_noerrors(response)
        res += response.content.decode('utf-8')
    jsn = self._conn._convert_to_json('[' + res + ']')
    return jsn[1:]

  def get_data_pandas(self, offset=0, length=None):
    """
    Returns a Pandas DataFrame containing frame data starting at a given
    offset and spanning a given length.

    Parameters
    ----------
    offset : int
      Position (index) of the first row to be retrieved.
      Optional. Default=0.
    length : int
      Maximum number of rows to be retrieved starting from the row
      indicated by offset. A value of 'None' means 'all rows' on and
      after the offset.
      Optional. Default=None.

    Returns
    -------
    Pandas DataFrame

    """
    import pandas
    # Uses json conversion which is slow, res can get huge
    # if the frame data is huge.
    responses = self._get_data(offset=offset, length=length)
    res = ''
    for response in responses:
        _assert_noerrors(response)
        res += response.content.decode('utf-8')
    jsn = self._conn._convert_to_json('[' + res + ']')
    return pandas.DataFrame(columns=jsn[0:1][0], data=jsn[1:])

  @property
  def num_rows(self):
    """int: Gets the number of rows of the table."""
    request = graph_proto.GetTableFramesRequest()
    request.name.extend([self._name])
    response = self._conn.call(request, self._conn._graph_svc.GetTableFrames)
    data = response.table_type[0]
    return data.nrows

  def _get_data(self, offset=0, length=None, headers=True,
                content_type=data_proto.JSON):
    if isinstance(offset, six.string_types):
      offset = int(offset)
    if isinstance(length, six.string_types):
      length = int(length)
    if isinstance(offset, int):
      if offset < 0:
        raise ValueError('offset is negative')
    if isinstance(length, int):
      if length < 0:
        raise ValueError('length is negative')

    request = data_proto.DownloadDataRequest()
    request.repository_name = self._name
    if offset is not None:
      request.offset.value = offset
    if length is not None:
      request.length.value = length
    request.with_headers = headers
    request.content_type = content_type
    responses = self._conn.call(request, self._conn._data_svc.DownloadData)
    return responses

  def _create_csv_packet (self, repository_name, data,
                          headerMode=HeaderMode.NONE):
    request = data_proto.UploadDataRequest()
    request.repository_name = repository_name.encode('utf-8')
    request.content = _to_bytes(data)
    request.content_type = data_proto.CSV

    if headerMode == HeaderMode.IGNORE:
      request.header_mode = data_proto.IGNORE_HEADERS
    elif headerMode == HeaderMode.NORMAL:
      request.header_mode = data_proto.NORMAL
    elif headerMode == HeaderMode.STRICT:
      request.header_mode = data_proto.STRICT
    else:
      request.header_mode = data_proto.NONE

    request.expected_server_id = self._conn._server_unique_identifier

    return request

  def _insert_csv_packet_generator(self, paths, headerMode):
    for fpath in paths:
        header = ''
        data = ''
        dsize = 0
        with open(fpath, 'rb') as f:
          line = f.readline()
          while line:
            line = _to_unicode(line)
            lsize = len(line)
            if (dsize + lsize) < MAX_PACKET_SIZE:
              data += line
              dsize += lsize
            else:
              yield self._create_csv_packet(self._name, data, headerMode)
              data = line
              dsize = len(data)
            line = f.readline()
          if data:
            yield self._create_csv_packet(self._name, data, headerMode)

  def _insert_from_csv(self, paths, headerMode=HeaderMode.NONE):
    data_iter = self._insert_csv_packet_generator(paths, headerMode)
    response = self._conn.call(data_iter, self._conn._data_svc.UploadData)

  def _ingest(self, paths, headerMode=HeaderMode.NONE):
    request = data_proto.IngestUriRequest()
    request.repository_name = self._name

    if isinstance(paths, (list, tuple)):
      request.content_uri.extend(paths)
    else:
      request.content_uri.extend([paths])

    if headerMode == HeaderMode.IGNORE:
      request.header_mode = data_proto.IGNORE_HEADERS
    elif headerMode == HeaderMode.NORMAL:
      request.header_mode = data_proto.NORMAL
    elif headerMode == HeaderMode.STRICT:
      request.header_mode = data_proto.STRICT
    else:
      request.header_mode = data_proto.NONE

    if (len(self._conn.aws_access_key_id) > 0 and \
        len(self._conn.aws_secret_access_key) > 0):
      request.authorization = self._conn.aws_access_key_id + ':' + \
                              self._conn.aws_secret_access_key

    response = self._conn.call(request, self._conn._data_svc.IngestUri)

  def _save_to_csv(self, path, offset=0, length=None, headers=False):
    # This will stream the bytes directly which is > 10X faster than using json.
    responses = self._get_data(offset=offset, length=length, headers=headers,
                               content_type=data_proto.CSV)
    with open(path, 'wb') as fobject:
      if headers:
        columns = [i[0] for i in self.schema]
        cnt = 0
        for col in columns:
          if cnt > 0:
            fobject.write(_to_bytes(','))
          cnt += 1
          fobject.write(_to_bytes(col))
        fobject.write('\n')
      # Each packet can be directly written to the file since we have the
      # raw data. This avoids extra conversion issues and extra memory from
      # json.
      for response in responses:
          _assert_noerrors(response)
          fobject.write(response.content)
    fobject.close()

  def _egest(self, path, headers=False):
    request = data_proto.EgestUriRequest()
    request.repository_name = self._name
    request.content_uri = path
    request.with_headers = headers
    response = self._conn.call(request, self._conn._data_svc.EgestUri)

  def __str__(self):
    print_frame = "{'name': '" + self.name + "'" + \
                  ", 'schema': " + _to_unicode(self.schema) + "}"
    return _unicode_to_str(print_frame)

# -----------------------------------------------------------------------------

class VertexFrame(TableFrame):
  """
  VertexFrame object represents a collection of vertices held on the xGT
  server; it can be used to retrieve information about them and should not be
  instantiated directly by the user. Methods that return this object:
  `Connection.get_vertex_frame()`, `Connection.get_vertex_frames()` and
  `Connection.create_vertex_frame()`.

  Each vertex in a VertexFrame shares the same properties,
  described in `VertexFrame.schema`. Each vertex in a VertexFrame
  is uniquely identified by the property listed in `VertexFrame.key`.

  Parameters
  ----------
  conn : Connection
    An open connection to an xGT server.
  name : str
    Name of vertex frame.
  schema : list of pairs
    List of pairs associating property names with xGT data types.
    Each vertex in the VertexFrame will have these properties.
  key : str
    The property name used to uniquely identify vertices in the graph.
    This is the name of one of the properties from the schema and
    must be unique for each vertex in the frame.
  error_frame_name : str
    Name of the frame created to store errors encountered while loading data.
    (internal)

  Examples
  --------
  >>> import xgt
  >>> conn = xgt.Connection()
  >>> v1 = conn.create_vertex_frame(
  ...        name = 'Person',
  ...        schema = [['id', xgt.INT],
  ...                  ['name', xgt.TEXT]],
  ...        key = 'id')
  >>> v2 = conn.get_vertex_frame('Company') #Company is an existing vertex frame
  >>> print(v1.name)

  """
  def __init__(self, conn, name, schema, key, error_frame_name):
    super(VertexFrame, self).__init__(conn, name, schema, error_frame_name)
    self._key = key

  @property
  def key(self):
    """str: Gets the property name that uniquely identifies vertices of this type."""
    return self._key

  @property
  def num_vertices(self):
    """int: Gets the number of vertices in the VertexFrame."""
    request = graph_proto.GetVertexFramesRequest()
    request.name.extend([self._name])
    response = self._conn.call(request, self._conn._graph_svc.GetVertexFrames)
    data = response.vertex_type[0]
    return data.nvertices

  # Override num_rows
  num_rows = num_vertices

  def __str__(self):
    print_frame = ("{'name': '" + self.name + "'" +
                   ", 'schema': " + _to_unicode(self.schema) +
                   ", 'key': '" + self.key + "'}")
    return _unicode_to_str(print_frame)

# -----------------------------------------------------------------------------

class EdgeFrame(TableFrame):
  """
  EdgeFrame object represents a collection of edges held on the xGT server;
  it can be used to retrieve information about them and should not be
  instantiated directly by the user. Methods that return this object:
  `Connection.get_edge_frame()`, `Connection.get_edge_frames()` and
  `Connection.create_edge_frame()`. Each edge in an EdgeFrame shares the same
  properties, described in `EdgeFrame.schema`.

  The source vertex of each edge in an EdgeFrame must belong to the same
  VertexFrame. This name of this VertexFrame is given by `EdgeFrame.source_name`.
  The targe vertex of each edge in an EdgeFrame must belong to the same
  VertexFrame. This name of this VertexFrame is given by `EdgeFrame.target_name`.

  For each edge in the EdgeFrame, its source vertex is identified by
  the edge property name given by `EdgeFrame.source_key`, which is
  be one of the properties listed in the schema. The edge target vertex
  is identified by the property name given by `EdgeFrame.target_key`.

  Parameters
  ----------
  conn : Connection
    An open connection to an xGT server.
  name : str
    Name of edge frame.
  schema : list of pairs
    List of pairs associating property names with xGT data types.
    Each edge in the EdgeFrame will have these properties.
  source : str or VertexFrame
    The name of a VertexFrame or a VertexFrame object.
    The source vertex of each edge in this EdgeFrame will belong
    to this VertexFrame.
  target : str or VertexFrame
    The name of a VertexFrame or a VertexFrame object.
    The target vertex of each edge in this EdgeFrame will belong
    to this VertexFrame.
  source_key : str
    The edge property name that identifies the source vertex of an edge.
    This is one of the properties from the schema.
  target_key : str
    The edge property name that identifies the target vertex of an edge.
    This is one of the properties from the schema.
  error_frame_name : str
    Name of the frame created to store errors encountered while loading data.
    (internal)

  Examples
  --------
  >>> import xgt
  >>> conn = xgt.Connection()
  >>> e1 = conn.create_edge_frame(
  ...        name = 'WorksFor',
  ...        schema = [['srcid', xgt.INT],
  ...                  ['role', xgt.TEXT],
  ...                  ['trgid', xgt.INT]],
  ...        source = 'Person',
  ...        target = 'Company',
  ...        source_key = 'srcid',
  ...        target_key = 'trgid')
  >>> e2 = conn.get_edge_frame('RelatedTo') #RelatedTo is an existing edge frame
  >>> print(e1.name)

  """
  def __init__(self, conn, name, schema, source, target, source_key, target_key, error_frame_name):
    super(EdgeFrame, self).__init__(conn, name, schema, error_frame_name)
    self._source_name = source
    self._target_name = target
    self._source_key = source_key
    self._target_key = target_key

  @property
  def source_name(self):
    """str: Gets the name of the source vertex frame."""
    return self._source_name

  @property
  def target_name(self):
    """str: Gets the name of the target vertex frame."""
    return self._target_name

  @property
  def source_key(self):
    """str: The edge property name that identifies the source vertex of an edge."""
    return self._source_key

  @property
  def target_key(self):
    """str: The edge property name that identifies the target vertex of an edge."""
    return self._target_key

  @property
  def num_edges(self):
    """int: Gets the number of edges of the edge frame."""
    request = graph_proto.GetEdgeFramesRequest()
    request.name.extend([self._name])
    response = self._conn.call(request, self._conn._graph_svc.GetEdgeFrames)
    data = response.edge_type[0]
    return data.nedges

  # Override num_rows
  num_rows = num_edges

  def __str__(self):
    print_frame = ("{'name': '" + self.name + "'" +
                   ", 'source': '" + self.source_name + "'" +
                   ", 'target': '" + self.target_name + "'" +
                   ", 'schema': " + _to_unicode(self.schema) +
                   ", 'source_key': '" + self.source_key + "'" +
                   ", 'target_key': '" + self.target_key + "'}")
    return _unicode_to_str(print_frame)

# -----------------------------------------------------------------------------

def _group_paths(paths, for_ingest):
  client_prefix = 'xgt://'
  server_prefix = 'xgtd://'
  url_prefixes = ['s3://', 'https://', 'http://', 'ftps://', 'ftp://']
  client_paths = []
  server_paths = []
  url_paths = []
  if isinstance(paths, six.string_types):
    paths = [paths]
  elif not isinstance(paths, (list, tuple)):
    return client_paths, server_paths, url_paths
  for one_path in paths:
    if one_path == "":
      raise ValueError('one or more "paths" are empty')
    if one_path.startswith(client_prefix):
      _validate_client_path(one_path)
      client_paths.append(one_path[len(client_prefix):])
    elif one_path.startswith(server_prefix):
      server_paths.append(one_path[len(server_prefix):])
    elif any(map(lambda p: one_path.startswith(p), url_prefixes)):
      for url_prefix in url_prefixes:
        if for_ingest == False:
          msg = 'Url paths are invalid for data writing ' \
                '"{0}".'.format(one_path)
          raise XgtNotImplemented(msg)
        if one_path.startswith(url_prefix):
          url_paths.append(one_path)
          break
    else:
      if '://' in one_path:
        msg = 'Unsupported url protocol in path "{0}".'.format(one_path)
        raise XgtNotImplemented(msg)
      _validate_client_path(one_path)
      client_paths.append(one_path)
  return client_paths, server_paths, url_paths

def _validate_client_path(one_path):
  if one_path.endswith('.gz') or one_path.endswith('.bz2'):
    msg = 'Loading compressed files from a local filesystem is ' \
          'not supported: {0}'.format(one_path)
    raise XgtNotImplemented(msg)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: GraphTypesService.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import ErrorMessages_pb2 as ErrorMessages__pb2
from . import SchemaMessages_pb2 as SchemaMessages__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='GraphTypesService.proto',
  package='xgt',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x17GraphTypesService.proto\x12\x03xgt\x1a\x13\x45rrorMessages.proto\x1a\x14SchemaMessages.proto\"]\n\x18\x43reateVertexFrameRequest\x12\x1a\n\x12\x65xpected_server_id\x18\x01 \x01(\x04\x12%\n\x0bvertex_type\x18\x02 \x01(\x0b\x32\x10.xgt.VertexFrame\"B\n\x16GetVertexFramesRequest\x12\x1a\n\x12\x65xpected_server_id\x18\x01 \x01(\x04\x12\x0c\n\x04name\x18\x02 \x03(\t\"[\n\x17GetVertexFramesResponse\x12\x19\n\x05\x65rror\x18\x01 \x03(\x0b\x32\n.xgt.Error\x12%\n\x0bvertex_type\x18\x02 \x03(\x0b\x32\x10.xgt.VertexFrame\"W\n\x16\x43reateEdgeFrameRequest\x12\x1a\n\x12\x65xpected_server_id\x18\x01 \x01(\x04\x12!\n\tedge_type\x18\x02 \x01(\x0b\x32\x0e.xgt.EdgeFrame\"@\n\x14GetEdgeFramesRequest\x12\x1a\n\x12\x65xpected_server_id\x18\x01 \x01(\x04\x12\x0c\n\x04name\x18\x02 \x03(\t\"U\n\x15GetEdgeFramesResponse\x12\x19\n\x05\x65rror\x18\x01 \x03(\x0b\x32\n.xgt.Error\x12!\n\tedge_type\x18\x02 \x03(\x0b\x32\x0e.xgt.EdgeFrame\"Z\n\x17\x43reateTableFrameRequest\x12\x1a\n\x12\x65xpected_server_id\x18\x01 \x01(\x04\x12#\n\ntable_type\x18\x02 \x01(\x0b\x32\x0f.xgt.TableFrame\"A\n\x15GetTableFramesRequest\x12\x1a\n\x12\x65xpected_server_id\x18\x01 \x01(\x04\x12\x0c\n\x04name\x18\x02 \x03(\t\"X\n\x16GetTableFramesResponse\x12\x19\n\x05\x65rror\x18\x01 \x03(\x0b\x32\n.xgt.Error\x12#\n\ntable_type\x18\x02 \x03(\x0b\x32\x0f.xgt.TableFrame\"=\n\x11\x44\x65leteTypeRequest\x12\x1a\n\x12\x65xpected_server_id\x18\x01 \x01(\x04\x12\x0c\n\x04name\x18\x02 \x01(\t\"J\n\x12\x44\x65leteTypeResponse\x12\x19\n\x05\x65rror\x18\x01 \x03(\x0b\x32\n.xgt.Error\x12\x19\n\x11\x66ound_and_deleted\x18\x02 \x01(\x08\x32\xae\x04\n\x11GraphTypesService\x12R\n\x11\x43reateVertexFrame\x12\x1d.xgt.CreateVertexFrameRequest\x1a\x1c.xgt.GetVertexFramesResponse\"\x00\x12N\n\x0fGetVertexFrames\x12\x1b.xgt.GetVertexFramesRequest\x1a\x1c.xgt.GetVertexFramesResponse\"\x00\x12L\n\x0f\x43reateEdgeFrame\x12\x1b.xgt.CreateEdgeFrameRequest\x1a\x1a.xgt.GetEdgeFramesResponse\"\x00\x12H\n\rGetEdgeFrames\x12\x19.xgt.GetEdgeFramesRequest\x1a\x1a.xgt.GetEdgeFramesResponse\"\x00\x12O\n\x10\x43reateTableFrame\x12\x1c.xgt.CreateTableFrameRequest\x1a\x1b.xgt.GetTableFramesResponse\"\x00\x12K\n\x0eGetTableFrames\x12\x1a.xgt.GetTableFramesRequest\x1a\x1b.xgt.GetTableFramesResponse\"\x00\x12?\n\nDeleteType\x12\x16.xgt.DeleteTypeRequest\x1a\x17.xgt.DeleteTypeResponse\"\x00\x62\x06proto3')
  ,
  dependencies=[ErrorMessages__pb2.DESCRIPTOR,SchemaMessages__pb2.DESCRIPTOR,])




_CREATEVERTEXFRAMEREQUEST = _descriptor.Descriptor(
  name='CreateVertexFrameRequest',
  full_name='xgt.CreateVertexFrameRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='expected_server_id', full_name='xgt.CreateVertexFrameRequest.expected_server_id', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='vertex_type', full_name='xgt.CreateVertexFrameRequest.vertex_type', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=75,
  serialized_end=168,
)


_GETVERTEXFRAMESREQUEST = _descriptor.Descriptor(
  name='GetVertexFramesRequest',
  full_name='xgt.GetVertexFramesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='expected_server_id', full_name='xgt.GetVertexFramesRequest.expected_server_id', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='xgt.GetVertexFramesRequest.name', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=170,
  serialized_end=236,
)


_GETVERTEXFRAMESRESPONSE = _descriptor.Descriptor(
  name='GetVertexFramesResponse',
  full_name='xgt.GetVertexFramesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='error', full_name='xgt.GetVertexFramesResponse.error', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='vertex_type', full_name='xgt.GetVertexFramesResponse.vertex_type', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=238,
  serialized_end=329,
)


_CREATEEDGEFRAMEREQUEST = _descriptor.Descriptor(
  name='CreateEdgeFrameRequest',
  full_name='xgt.CreateEdgeFrameRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='expected_server_id', full_name='xgt.CreateEdgeFrameRequest.expected_server_id', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='edge_type', full_name='xgt.CreateEdgeFrameRequest.edge_type', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=331,
  serialized_end=418,
)


_GETEDGEFRAMESREQUEST = _descriptor.Descriptor(
  name='GetEdgeFramesRequest',
  full_name='xgt.GetEdgeFramesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='expected_server_id', full_name='xgt.GetEdgeFramesRequest.expected_server_id', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='xgt.GetEdgeFramesRequest.name', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=420,
  serialized_end=484,
)


_GETEDGEFRAMESRESPONSE = _descriptor.Descriptor(
  name='GetEdgeFramesResponse',
  full_name='xgt.GetEdgeFramesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='error', full_name='xgt.GetEdgeFramesResponse.error', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='edge_type', full_name='xgt.GetEdgeFramesResponse.edge_type', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=486,
  serialized_end=571,
)


_CREATETABLEFRAMEREQUEST = _descriptor.Descriptor(
  name='CreateTableFrameRequest',
  full_name='xgt.CreateTableFrameRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='expected_server_id', full_name='xgt.CreateTableFrameRequest.expected_server_id', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='table_type', full_name='xgt.CreateTableFrameRequest.table_type', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=573,
  serialized_end=663,
)


_GETTABLEFRAMESREQUEST = _descriptor.Descriptor(
  name='GetTableFramesRequest',
  full_name='xgt.GetTableFramesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='expected_server_id', full_name='xgt.GetTableFramesRequest.expected_server_id', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='xgt.GetTableFramesRequest.name', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=665,
  serialized_end=730,
)


_GETTABLEFRAMESRESPONSE = _descriptor.Descriptor(
  name='GetTableFramesResponse',
  full_name='xgt.GetTableFramesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='error', full_name='xgt.GetTableFramesResponse.error', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='table_type', full_name='xgt.GetTableFramesResponse.table_type', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=732,
  serialized_end=820,
)


_DELETETYPEREQUEST = _descriptor.Descriptor(
  name='DeleteTypeRequest',
  full_name='xgt.DeleteTypeRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='expected_server_id', full_name='xgt.DeleteTypeRequest.expected_server_id', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='xgt.DeleteTypeRequest.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=822,
  serialized_end=883,
)


_DELETETYPERESPONSE = _descriptor.Descriptor(
  name='DeleteTypeResponse',
  full_name='xgt.DeleteTypeResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='error', full_name='xgt.DeleteTypeResponse.error', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='found_and_deleted', full_name='xgt.DeleteTypeResponse.found_and_deleted', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=885,
  serialized_end=959,
)

_CREATEVERTEXFRAMEREQUEST.fields_by_name['vertex_type'].message_type = SchemaMessages__pb2._VERTEXFRAME
_GETVERTEXFRAMESRESPONSE.fields_by_name['error'].message_type = ErrorMessages__pb2._ERROR
_GETVERTEXFRAMESRESPONSE.fields_by_name['vertex_type'].message_type = SchemaMessages__pb2._VERTEXFRAME
_CREATEEDGEFRAMEREQUEST.fields_by_name['edge_type'].message_type = SchemaMessages__pb2._EDGEFRAME
_GETEDGEFRAMESRESPONSE.fields_by_name['error'].message_type = ErrorMessages__pb2._ERROR
_GETEDGEFRAMESRESPONSE.fields_by_name['edge_type'].message_type = SchemaMessages__pb2._EDGEFRAME
_CREATETABLEFRAMEREQUEST.fields_by_name['table_type'].message_type = SchemaMessages__pb2._TABLEFRAME
_GETTABLEFRAMESRESPONSE.fields_by_name['error'].message_type = ErrorMessages__pb2._ERROR
_GETTABLEFRAMESRESPONSE.fields_by_name['table_type'].message_type = SchemaMessages__pb2._TABLEFRAME
_DELETETYPERESPONSE.fields_by_name['error'].message_type = ErrorMessages__pb2._ERROR
DESCRIPTOR.message_types_by_name['CreateVertexFrameRequest'] = _CREATEVERTEXFRAMEREQUEST
DESCRIPTOR.message_types_by_name['GetVertexFramesRequest'] = _GETVERTEXFRAMESREQUEST
DESCRIPTOR.message_types_by_name['GetVertexFramesResponse'] = _GETVERTEXFRAMESRESPONSE
DESCRIPTOR.message_types_by_name['CreateEdgeFrameRequest'] = _CREATEEDGEFRAMEREQUEST
DESCRIPTOR.message_types_by_name['GetEdgeFramesRequest'] = _GETEDGEFRAMESREQUEST
DESCRIPTOR.message_types_by_name['GetEdgeFramesResponse'] = _GETEDGEFRAMESRESPONSE
DESCRIPTOR.message_types_by_name['CreateTableFrameRequest'] = _CREATETABLEFRAMEREQUEST
DESCRIPTOR.message_types_by_name['GetTableFramesRequest'] = _GETTABLEFRAMESREQUEST
DESCRIPTOR.message_types_by_name['GetTableFramesResponse'] = _GETTABLEFRAMESRESPONSE
DESCRIPTOR.message_types_by_name['DeleteTypeRequest'] = _DELETETYPEREQUEST
DESCRIPTOR.message_types_by_name['DeleteTypeResponse'] = _DELETETYPERESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CreateVertexFrameRequest = _reflection.GeneratedProtocolMessageType('CreateVertexFrameRequest', (_message.Message,), dict(
  DESCRIPTOR = _CREATEVERTEXFRAMEREQUEST,
  __module__ = 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.CreateVertexFrameRequest)
  ))
_sym_db.RegisterMessage(CreateVertexFrameRequest)

GetVertexFramesRequest = _reflection.GeneratedProtocolMessageType('GetVertexFramesRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETVERTEXFRAMESREQUEST,
  __module__ = 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.GetVertexFramesRequest)
  ))
_sym_db.RegisterMessage(GetVertexFramesRequest)

GetVertexFramesResponse = _reflection.GeneratedProtocolMessageType('GetVertexFramesResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETVERTEXFRAMESRESPONSE,
  __module__ = 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.GetVertexFramesResponse)
  ))
_sym_db.RegisterMessage(GetVertexFramesResponse)

CreateEdgeFrameRequest = _reflection.GeneratedProtocolMessageType('CreateEdgeFrameRequest', (_message.Message,), dict(
  DESCRIPTOR = _CREATEEDGEFRAMEREQUEST,
  __module__ = 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.CreateEdgeFrameRequest)
  ))
_sym_db.RegisterMessage(CreateEdgeFrameRequest)

GetEdgeFramesRequest = _reflection.GeneratedProtocolMessageType('GetEdgeFramesRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETEDGEFRAMESREQUEST,
  __module__ = 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.GetEdgeFramesRequest)
  ))
_sym_db.RegisterMessage(GetEdgeFramesRequest)

GetEdgeFramesResponse = _reflection.GeneratedProtocolMessageType('GetEdgeFramesResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETEDGEFRAMESRESPONSE,
  __module__ = 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.GetEdgeFramesResponse)
  ))
_sym_db.RegisterMessage(GetEdgeFramesResponse)

CreateTableFrameRequest = _reflection.GeneratedProtocolMessageType('CreateTableFrameRequest', (_message.Message,), dict(
  DESCRIPTOR = _CREATETABLEFRAMEREQUEST,
  __module__ = 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.CreateTableFrameRequest)
  ))
_sym_db.RegisterMessage(CreateTableFrameRequest)

GetTableFramesRequest = _reflection.GeneratedProtocolMessageType('GetTableFramesRequest', (_message.Message,), dict(
  DESCRIPTOR = _GETTABLEFRAMESREQUEST,
  __module__ = 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.GetTableFramesRequest)
  ))
_sym_db.RegisterMessage(GetTableFramesRequest)

GetTableFramesResponse = _reflection.GeneratedProtocolMessageType('GetTableFramesResponse', (_message.Message,), dict(
  DESCRIPTOR = _GETTABLEFRAMESRESPONSE,
  __module__ = 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.GetTableFramesResponse)
  ))
_sym_db.RegisterMessage(GetTableFramesResponse)

DeleteTypeRequest = _reflection.GeneratedProtocolMessageType('DeleteTypeRequest', (_message.Message,), dict(
  DESCRIPTOR = _DELETETYPEREQUEST,
  __module__ = 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.DeleteTypeRequest)
  ))
_sym_db.RegisterMessage(DeleteTypeRequest)

DeleteTypeResponse = _reflection.GeneratedProtocolMessageType('DeleteTypeResponse', (_message.Message,), dict(
  DESCRIPTOR = _DELETETYPERESPONSE,
  __module__ = 'GraphTypesService_pb2'
  # @@protoc_insertion_point(class_scope:xgt.DeleteTypeResponse)
  ))
_sym_db.RegisterMessage(DeleteTypeResponse)



_GRAPHTYPESSERVICE = _descriptor.ServiceDescriptor(
  name='GraphTypesService',
  full_name='xgt.GraphTypesService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=962,
  serialized_end=1520,
  methods=[
  _descriptor.MethodDescriptor(
    name='CreateVertexFrame',
    full_name='xgt.GraphTypesService.CreateVertexFrame',
    index=0,
    containing_service=None,
    input_type=_CREATEVERTEXFRAMEREQUEST,
    output_type=_GETVERTEXFRAMESRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetVertexFrames',
    full_name='xgt.GraphTypesService.GetVertexFrames',
    index=1,
    containing_service=None,
    input_type=_GETVERTEXFRAMESREQUEST,
    output_type=_GETVERTEXFRAMESRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='CreateEdgeFrame',
    full_name='xgt.GraphTypesService.CreateEdgeFrame',
    index=2,
    containing_service=None,
    input_type=_CREATEEDGEFRAMEREQUEST,
    output_type=_GETEDGEFRAMESRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetEdgeFrames',
    full_name='xgt.GraphTypesService.GetEdgeFrames',
    index=3,
    containing_service=None,
    input_type=_GETEDGEFRAMESREQUEST,
    output_type=_GETEDGEFRAMESRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='CreateTableFrame',
    full_name='xgt.GraphTypesService.CreateTableFrame',
    index=4,
    containing_service=None,
    input_type=_CREATETABLEFRAMEREQUEST,
    output_type=_GETTABLEFRAMESRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetTableFrames',
    full_name='xgt.GraphTypesService.GetTableFrames',
    index=5,
    containing_service=None,
    input_type=_GETTABLEFRAMESREQUEST,
    output_type=_GETTABLEFRAMESRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteType',
    full_name='xgt.GraphTypesService.DeleteType',
    index=6,
    containing_service=None,
    input_type=_DELETETYPEREQUEST,
    output_type=_DELETETYPERESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_GRAPHTYPESSERVICE)

DESCRIPTOR.services_by_name['GraphTypesService'] = _GRAPHTYPESSERVICE

# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import GraphTypesService_pb2 as GraphTypesService__pb2


class GraphTypesServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.CreateVertexFrame = channel.unary_unary(
        '/xgt.GraphTypesService/CreateVertexFrame',
        request_serializer=GraphTypesService__pb2.CreateVertexFrameRequest.SerializeToString,
        response_deserializer=GraphTypesService__pb2.GetVertexFramesResponse.FromString,
        )
    self.GetVertexFrames = channel.unary_unary(
        '/xgt.GraphTypesService/GetVertexFrames',
        request_serializer=GraphTypesService__pb2.GetVertexFramesRequest.SerializeToString,
        response_deserializer=GraphTypesService__pb2.GetVertexFramesResponse.FromString,
        )
    self.CreateEdgeFrame = channel.unary_unary(
        '/xgt.GraphTypesService/CreateEdgeFrame',
        request_serializer=GraphTypesService__pb2.CreateEdgeFrameRequest.SerializeToString,
        response_deserializer=GraphTypesService__pb2.GetEdgeFramesResponse.FromString,
        )
    self.GetEdgeFrames = channel.unary_unary(
        '/xgt.GraphTypesService/GetEdgeFrames',
        request_serializer=GraphTypesService__pb2.GetEdgeFramesRequest.SerializeToString,
        response_deserializer=GraphTypesService__pb2.GetEdgeFramesResponse.FromString,
        )
    self.CreateTableFrame = channel.unary_unary(
        '/xgt.GraphTypesService/CreateTableFrame',
        request_serializer=GraphTypesService__pb2.CreateTableFrameRequest.SerializeToString,
        response_deserializer=GraphTypesService__pb2.GetTableFramesResponse.FromString,
        )
    self.GetTableFrames = channel.unary_unary(
        '/xgt.GraphTypesService/GetTableFrames',
        request_serializer=GraphTypesService__pb2.GetTableFramesRequest.SerializeToString,
        response_deserializer=GraphTypesService__pb2.GetTableFramesResponse.FromString,
        )
    self.DeleteType = channel.unary_unary(
        '/xgt.GraphTypesService/DeleteType',
        request_serializer=GraphTypesService__pb2.DeleteTypeRequest.SerializeToString,
        response_deserializer=GraphTypesService__pb2.DeleteTypeResponse.FromString,
        )


class GraphTypesServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def CreateVertexFrame(self, request, context):
    """Vertex frames
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetVertexFrames(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def CreateEdgeFrame(self, request, context):
    """Edge frames
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetEdgeFrames(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def CreateTableFrame(self, request, context):
    """Table frames
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTableFrames(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DeleteType(self, request, context):
    """Multi-type functions
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_GraphTypesServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'CreateVertexFrame': grpc.unary_unary_rpc_method_handler(
          servicer.CreateVertexFrame,
          request_deserializer=GraphTypesService__pb2.CreateVertexFrameRequest.FromString,
          response_serializer=GraphTypesService__pb2.GetVertexFramesResponse.SerializeToString,
      ),
      'GetVertexFrames': grpc.unary_unary_rpc_method_handler(
          servicer.GetVertexFrames,
          request_deserializer=GraphTypesService__pb2.GetVertexFramesRequest.FromString,
          response_serializer=GraphTypesService__pb2.GetVertexFramesResponse.SerializeToString,
      ),
      'CreateEdgeFrame': grpc.unary_unary_rpc_method_handler(
          servicer.CreateEdgeFrame,
          request_deserializer=GraphTypesService__pb2.CreateEdgeFrameRequest.FromString,
          response_serializer=GraphTypesService__pb2.GetEdgeFramesResponse.SerializeToString,
      ),
      'GetEdgeFrames': grpc.unary_unary_rpc_method_handler(
          servicer.GetEdgeFrames,
          request_deserializer=GraphTypesService__pb2.GetEdgeFramesRequest.FromString,
          response_serializer=GraphTypesService__pb2.GetEdgeFramesResponse.SerializeToString,
      ),
      'CreateTableFrame': grpc.unary_unary_rpc_method_handler(
          servicer.CreateTableFrame,
          request_deserializer=GraphTypesService__pb2.CreateTableFrameRequest.FromString,
          response_serializer=GraphTypesService__pb2.GetTableFramesResponse.SerializeToString,
      ),
      'GetTableFrames': grpc.unary_unary_rpc_method_handler(
          servicer.GetTableFrames,
          request_deserializer=GraphTypesService__pb2.GetTableFramesRequest.FromString,
          response_serializer=GraphTypesService__pb2.GetTableFramesResponse.SerializeToString,
      ),
      'DeleteType': grpc.unary_unary_rpc_method_handler(
          servicer.DeleteType,
          request_deserializer=GraphTypesService__pb2.DeleteTypeRequest.FromString,
          response_serializer=GraphTypesService__pb2.DeleteTypeResponse.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'xgt.GraphTypesService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# -*- coding: utf-8 -*- --------------------------------------------------===#
#
#  Copyright 2018-2019 Trovares Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and 
#  limitations under the License.
#
#===----------------------------------------------------------------------===#

import datetime as dt
import json
import logging
import math
import re
import sys

import grpc
import six

from . import ErrorMessages_pb2 as err_proto
from . import SchemaMessages_pb2 as sch_proto
from . import AdminService_pb2 as admin_proto
from . import AdminService_pb2_grpc as admin_grpc
from . import JobService_pb2 as job_proto
from . import JobService_pb2_grpc as job_grpc
from . import GraphTypesService_pb2 as graptype_proto
from . import GraphTypesService_pb2_grpc as graptype_grpc

log = logging.getLogger(__name__)

BOOLEAN = 'boolean'
INT = 'int'
FLOAT = 'float'
DATE = 'date'
TIME = 'time'
DATETIME = 'datetime'
IPADDRESS = 'ipaddress'
TEXT = 'text'

DEFAULT_OPT_LEVEL = 2
# Send in 2MB chunks (grpc recommends 16-64 KB, but this got the best perf locally)
# FYI: by default grpc only supports up to 4MB.
MAX_PACKET_SIZE = 2097152

class XgtError(Exception):
  """
  Base exception class from which all other xgt exceptions inherit. It is
  raised in error cases that don't have a specific xgt exception type.
  """
  def __init__(self, msg, trace=''):
    if six.PY2:
      if isinstance(msg, unicode):
        msg = msg.encode('utf-8')
      if isinstance(trace, unicode):
        trace = trace.encode('utf-8')
    self.msg = msg
    self.trace = trace

    if log.getEffectiveLevel() >= logging.DEBUG:
      if self.trace != '':
        log.debug(self.trace)
      else:
        log.debug(self.msg)
    Exception.__init__(self, self.msg)

class XgtNotImplemented(XgtError):
  """Raised for functionality with pending implementation."""
class XgtInternalError(XgtError):
  """
  Intended for internal server purposes only. This exception should not become
  visible to the user.
  """
class XgtIOError(XgtError):
  """An I/O problem occured either on the client or server side."""
class XgtServerMemoryError(XgtError):
  """
  The server memory usage is close to or at capacity and work could be lost.
  """
class XgtConnectionError(XgtError):
  """
  The client cannot properly connect to the server. This can include a failure
  to connect due to an xgt module version error.
  """
class XgtSyntaxError(XgtError):
  """A query was provided with incorrect syntax."""
class XgtTypeError(XgtError):
  """
  An unexpected type was supplied.

  For queries, an invalid data type was used either as an entity or as a
  property. For frames, either an edge, vertex or table frames was expected
  but the wrong frame type or some other data type was provided. For
  properties, the property declaration establishes the expected data type. A
  type error is raise if the data type used is not appropriate.
  """
class XgtValueError(XgtError):
  """An invalid or unexpected value was provided."""
class XgtNameError(XgtError):
  """
  An unexpected name was provided. Typically can occur during object retrieval
  where the object name was not found.
  """
class XgtArithmeticError(XgtError):
  """An invalid arithmetic calculation was detected and cannot be handled."""
class XgtFrameDependencyError(XgtError):
  """
  The requested action will produce an invalid graph or break a valid graph.
  """
class XgtTransactionError(XgtError):
  """A Transaction was attempted but didn't complete."""

class Job(object):
  """
  Represents a user-scheduled Job.

  An instance of this object is created by job-scheduling functions like
  `xgt.Connection.run_job` and `xgt.Connection.schedule_job`.

  A `Job` is used as a proxy for a job in the server and allows the user
  to monitor its execution, possibly cancel it, and learn about its status
  during and after execution.

  """
  def __init__(self, conn, job_id):
    self._conn = conn
    self._id = job_id

  def _get_job_data(self):
    request = job_proto.GetJobsRequest()
    request.job_id.extend([self._id])
    response = self._conn.call(request, self._conn._job_svc.GetJobs)
    job = response.job_status[0]  # Retrieve only one job from the list.
    job_data = {
      'jobid': job.job_id,
      'status': job_proto.JobStatusEnum.Name(job.status).lower(),
      'start_time': job.start_time.ToDatetime().isoformat(),
      'end_time': job.end_time.ToDatetime().isoformat(),
      'error_type': None }
    if job.error:
      if len(job.error) > 0:
        error_code_name = err_proto.ErrorCodeEnum.Name(job.error[0].code)
        job_data['error_type'] = _code_error_map[error_code_name]
        job_data['error'] = ', '.join([e.message for e in job.error])
        job_data['trace'] = ', '.join([e.detail for e in job.error])

    if log.getEffectiveLevel() >= logging.DEBUG:
      job_id = job_data['jobid']
      job_status = job_data['status']
      if 'error' in job_data:
        error = job_data['error']
      else:
        error = ''
      if 'trace' in job_data:
        trace = job_data['trace']
      else:
        trace = ''
      msg = u'Job: {0} Status: {1}'.format(_to_unicode(job_id),
                                          job_status)
      if error != '':
        msg = msg + "\nError: \n" + error
      if trace != '':
        msg = msg + "\nTrace: \n" + trace
      log.debug(msg)

    return job_data

  @property
  def id(self):
    """
    int: Identifier of the job.

    A 64-bit integer value that uniquely identifies a job. It is
    automatically incremented for each scheduled job over the lifetime of
    the `xgtd` server process.

    """
    return self._id

  @property
  def status(self):
    """
    str: Status of the job.

        ============  ===============================================
        Job status
        -------------------------------------------------------------
           Status                       Description
        ============  ===============================================
           scheduled  The state after the job has been created, but
                      before it has started running.
             running  The job is being executed.
           completed  The job finished successfully.
            canceled  The job was canceled.
              failed  The job failed. When the job fails the `error`
                      and `trace` properties are populated.
        ============  ===============================================

    """
    data = self._get_job_data()
    if 'status' in data:
      return _to_unicode(data['status'])
    else:
      return ''

  @property
  def start_time(self):
    """
    str: Date and time when the job was scheduled.

    This is a formatted string that has a resolution of seconds.
    """
    data = self._get_job_data()
    if 'start_time' in data:
      return dt.datetime.strptime(data['start_time'], '%Y-%m-%dT%H:%M:%S')
    else:
      return ''

  @property
  def end_time(self):
    """
    str: Date and time when the job finished running.

    This is a formatted string that has a resolution of seconds.
    """
    data = self._get_job_data()
    if 'end_time' in data:
      return dt.datetime.strptime(data['end_time'], '%Y-%m-%dT%H:%M:%S')
    else:
      return ''

  @property
  def error_type(self):
    """
    object: Class that belongs to the XgtError hierarchy that corresponds to
            the original exception type thrown and caused the Job to fail.
    """
    data = self._get_job_data()
    if 'error_type' in data:
      return data['error_type']
    else:
      return XgtError

  @property
  def error(self):
    """
    str: User-friendly error message describing the reason a job failed.
    """
    data = self._get_job_data()
    if 'error' in data:
      return _to_unicode(data['error'])
    else:
      return ''

  @property
  def trace(self):
    """
    str: Very detailed error message for a failed job.

    This error message contains the friendly error message and a stack
    strace for the code that participated in the error.
    """
    data = self._get_job_data()
    if 'trace' in data:
      return _to_unicode(data['trace'])
    else:
      return ''

  def __str__(self):
    txt = (u'id:{0}, status:{1}').format(self.id, self.status)
    if len(self.error) > 0:
      txt = txt + (u', nerror:{0}').format(self.error)
    return _unicode_to_str(txt)

# Validation support functions

def _validated_schema(obj):
  '''Takes a user-supplied object and returns a valid schema.

  Users can supply a variety of objects as valid schemas. To simplify internal
  processing, we canonicalize these into a list of string-type pairs,
  performing validation along the way.
  '''
  # Validate the shape first
  try:
    if len(obj) < 1:
      raise XgtTypeError('A schema must not be empty.')
    for pairlike in obj:
      assert len(pairlike) == 2
  except:
    raise XgtTypeError('A schema must be a non-empty list of (property, type) pairs.')
  # Looks good. Return a canonical schema.
  return [(_validated_property_name(name), _validated_property_type(xgt_type))
          for name,xgt_type in obj]

def _validated_frame_name(obj):
  '''Takes a user-supplied object and returns a unicode frame name string.'''
  name = _as_unicode(obj)
  if len(name) < 1:
    raise XgtValueError('Frame names cannot be empty.')
  if u'.' in name:
    raise XgtValueError('Frame names cannot contain periods: '+name)
  return name

def _validated_property_name(obj):
  '''Takes a user-supplied object and returns a unicode proprty name string.'''
  name = _as_unicode(obj)
  return name

def _validated_property_type(obj):
  '''Takes a user-supplied object and returns an xGT schema type.'''
  prop_type = _as_unicode(obj)
  valid_prop_types = [BOOLEAN, INT, FLOAT, DATE, TIME, DATETIME, IPADDRESS, TEXT]
  if prop_type.lower() not in valid_prop_types:
    raise XgtTypeError('Invalid property type "'+prop_type+'"')
  return prop_type.upper()

def _validate_opt_level(optlevel):
  """
  Valid optimization level values are:
    - 0: No optimization.
    - 1: General optimization.
    - 2: WHERE-clause optimization.
    - 3: Degree-cycle optimization.
  """
  if isinstance(optlevel, int):
    if optlevel not in [0, 1, 2, 3]:
      raise XgtValueError("Invalid optlevel '" + str(optlevel) +"'")
  else:
    raise XgtTypeError("optlevel must be an integer")
  return True

def _assert_noerrors(response):
  if len(response.error) > 0:
    error = response.error[0]
    try:
      error_code_name = err_proto.ErrorCodeEnum.Name(error.code)
      error_class = _code_error_map[error_code_name]
      raise error_class(error.message, error.detail)
    except XgtError:
      raise
    except Exception as ex:
      raise XgtError("Error detected while raising exception" +
                     _to_unicode(ex), _to_unicode(ex))

def _assert_isstring(name, text):
  if not isinstance(text, six.string_types):
    msg = name + " is not a string: '" + _to_unicode(text) + "'"
    raise TypeError(msg)

def _obj_to_str(obj, level=0):
  ret = ''
  if isinstance(obj, (str, six.text_type)):
    ret = ret + six.text_type(obj)
  elif isinstance(obj, list):
    ret = ret + _list_to_str(obj, level)
  elif isinstance(obj, dict):
    ret = ret + _dict_to_str(obj, level)
  else:
    ret = ret + '\n' + json.dumps(obj, indent=4, sort_keys=True)
  return ret

def _dict_to_str(obj, level):
  ret = ''
  indent = '\n' + '    ' * level
  for k in list(obj):
    ret = ret + indent + six.text_type(k) + ': '
    ret = ret + _obj_to_str(obj[k], level+1)
  return ret

def _list_to_str(obj, level):
  ret = ''
  indent = '\n' + '    ' * level
  for i in obj:
    ret = ret + indent + _obj_to_str(i, level+1)
  return ret

# Unicode support functions.

# Why the two functions _to_* and _as_*? There are two general cases when you
# want to convert something to a certain type of string. The first case is if
# you are creating a string that will be handed to a user. Exceptions and log
# messages fall into this category. In this case, you really want a
# *constructor*, turning any object into some type of string, even if it
# doesn't natively contain character data. The second case is when you want to
# collapse unicode and byte data into one or the other. This includes most xGT
# operations, referencing frames or properties, for instance. In this case, if
# the method is passed some other type (maybe an int or class instance), you
# want an exception to be thrown.
#
# So, to summarize:
#   _to_unicode( u'xyzzy' ) => u'xyzzy'
#   _to_unicode( b'xyzzy' ) => u'xyzzy'
#   _to_unicode( 12345.67 ) => u'12345.67'
#
#   _as_unicode( u'xyzzy' ) => u'xyzzy'
#   _as_unicode( b'xyzzy' ) => u'xyzzy'
#   _as_unicode( 12345.67 ) => TypeError
#
# _to_bytes/_as_bytes and _to_str work similarly (_as_str is not provided, as
# it would effectively be the same as the str() built-in).

def _unicode_to_str(text):
  if six.PY2 and isinstance(text, unicode):
    return text.encode('utf-8')
  else:
    return text

def _to_unicode(value):
  '''Constructs a unicode string out of any object.'''
  if isinstance(value, six.text_type):
    return value
  elif isinstance(value, six.binary_type):
    return value.decode('utf-8')
  else:
    return six.text_type(value)

def _as_unicode(value):
  '''Converts unicode or binary strings to unicode. Other types raise a TypeError.'''
  if isinstance(value, six.text_type):
    return value
  elif isinstance(value, six.binary_type):
    return value.decode('utf-8')
  raise TypeError('Cannot convert '+six.text_type(type(value))+' to unicode.')

def _to_bytes(value):
  '''Constructs a human-readable byte string out of any object.'''
  if isinstance(value, six.text_type):
    return value.encode('utf-8')
  elif isinstance(value, six.binary_type):
    return value
  else:
    return six.text_type(value).encode('utf-8')

def _as_bytes(value):
  '''Converts unicode or binary strings to bytes. Other types raise a TypeError.'''
  if isinstance(value, six.text_type):
    return value.encode('utf-8')
  elif isinstance(value, six.binary_type):
    return value
  raise TypeError('Cannot convert '+six.text_type(type(value))+' to bytes.')

def _to_str(value):
  '''Constructs a str out of any object.'''
  # We can't use the str built-in as a constructor because Python 2 believes
  # that unicode characters cannot be converted to str types. We want to.
  if isinstance(value, str):
    return value
  elif six.PY2 and isinstance(value, unicode):
    return value.encode('utf-8')
  elif six.PY3 and isinstance(value, bytes):
    return value.decode('utf-8')
  else:
    return str(value)

_code_error_map = {
  'GENERIC_ERROR': XgtError,
  'NOT_IMPLEMENTED': XgtNotImplemented,
  'INTERNAL_ERROR': XgtInternalError,
  'IO_ERROR': XgtIOError,
  'SERVER_MEMORY_ERROR': XgtServerMemoryError,
  'CONNECTION_ERROR': XgtConnectionError,
  'SYNTAX_ERROR': XgtSyntaxError,
  'TYPE_ERROR': XgtTypeError,
  'VALUE_ERROR': XgtValueError,
  'NAME_ERROR': XgtNameError,
  'ARITHMETIC_ERROR': XgtArithmeticError,
  'FRAME_DEPENDENCY_ERROR': XgtFrameDependencyError,
  'TRANSACTION_ERROR': XgtTransactionError,
}



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: SchemaMessages.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='SchemaMessages.proto',
  package='xgt',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x14SchemaMessages.proto\x12\x03xgt\"[\n\x08Property\x12\x0c\n\x04name\x18\x01 \x01(\t\x12$\n\tdata_type\x18\x02 \x01(\x0e\x32\x11.xgt.UvalTypeEnum\x12\x1b\n\x04role\x18\x03 \x01(\x0e\x32\r.xgt.RoleEnum\")\n\x06Schema\x12\x1f\n\x08property\x18\x01 \x03(\x0b\x32\r.xgt.Property\"Z\n\nTableFrame\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1b\n\x06schema\x18\x02 \x01(\x0b\x32\x0b.xgt.Schema\x12\r\n\x05nrows\x18\x03 \x01(\x04\x12\x12\n\nerror_name\x18\x04 \x01(\t\"_\n\x0bVertexFrame\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tnvertices\x18\x02 \x01(\x04\x12\x1b\n\x06schema\x18\x03 \x01(\x0b\x32\x0b.xgt.Schema\x12\x12\n\nerror_name\x18\x04 \x01(\t\"\xb0\x01\n\tEdgeFrame\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06nedges\x18\x02 \x01(\x04\x12\x1b\n\x06schema\x18\x03 \x01(\x0b\x32\x0b.xgt.Schema\x12\x15\n\rsource_vertex\x18\x04 \x01(\t\x12\x15\n\rtarget_vertex\x18\x05 \x01(\t\x12\x12\n\nsource_key\x18\x06 \x01(\t\x12\x12\n\ntarget_key\x18\x07 \x01(\t\x12\x12\n\nerror_name\x18\x08 \x01(\t*\x8e\x01\n\x0cUvalTypeEnum\x12\x15\n\x11UNKNOWN_UVAL_TYPE\x10\x00\x12\x0b\n\x07UNKNOWN\x10\x01\x12\x0b\n\x07\x42OOLEAN\x10\x02\x12\x07\n\x03INT\x10\x03\x12\t\n\x05\x46LOAT\x10\x04\x12\x08\n\x04\x44\x41TE\x10\x05\x12\x08\n\x04TIME\x10\x06\x12\x0c\n\x08\x44\x41TETIME\x10\x07\x12\r\n\tIPADDRESS\x10\x08\x12\x08\n\x04TEXT\x10\t*R\n\x08RoleEnum\x12\x0c\n\x08PROPERTY\x10\x00\x12\x0e\n\nVERTEX_KEY\x10\x01\x12\x13\n\x0f\x45\x44GE_SOURCE_KEY\x10\x02\x12\x13\n\x0f\x45\x44GE_TARGET_KEY\x10\x03\x62\x06proto3')
)

_UVALTYPEENUM = _descriptor.EnumDescriptor(
  name='UvalTypeEnum',
  full_name='xgt.UvalTypeEnum',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN_UVAL_TYPE', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNKNOWN', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='BOOLEAN', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INT', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLOAT', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATE', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TIME', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATETIME', index=7, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='IPADDRESS', index=8, number=8,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TEXT', index=9, number=9,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=534,
  serialized_end=676,
)
_sym_db.RegisterEnumDescriptor(_UVALTYPEENUM)

UvalTypeEnum = enum_type_wrapper.EnumTypeWrapper(_UVALTYPEENUM)
_ROLEENUM = _descriptor.EnumDescriptor(
  name='RoleEnum',
  full_name='xgt.RoleEnum',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='PROPERTY', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='VERTEX_KEY', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EDGE_SOURCE_KEY', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EDGE_TARGET_KEY', index=3, number=3,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=678,
  serialized_end=760,
)
_sym_db.RegisterEnumDescriptor(_ROLEENUM)

RoleEnum = enum_type_wrapper.EnumTypeWrapper(_ROLEENUM)
UNKNOWN_UVAL_TYPE = 0
UNKNOWN = 1
BOOLEAN = 2
INT = 3
FLOAT = 4
DATE = 5
TIME = 6
DATETIME = 7
IPADDRESS = 8
TEXT = 9
PROPERTY = 0
VERTEX_KEY = 1
EDGE_SOURCE_KEY = 2
EDGE_TARGET_KEY = 3



_PROPERTY = _descriptor.Descriptor(
  name='Property',
  full_name='xgt.Property',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='xgt.Property.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data_type', full_name='xgt.Property.data_type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='role', full_name='xgt.Property.role', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29,
  serialized_end=120,
)


_SCHEMA = _descriptor.Descriptor(
  name='Schema',
  full_name='xgt.Schema',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='property', full_name='xgt.Schema.property', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=122,
  serialized_end=163,
)


_TABLEFRAME = _descriptor.Descriptor(
  name='TableFrame',
  full_name='xgt.TableFrame',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='xgt.TableFrame.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='schema', full_name='xgt.TableFrame.schema', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='nrows', full_name='xgt.TableFrame.nrows', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error_name', full_name='xgt.TableFrame.error_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=165,
  serialized_end=255,
)


_VERTEXFRAME = _descriptor.Descriptor(
  name='VertexFrame',
  full_name='xgt.VertexFrame',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='xgt.VertexFrame.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='nvertices', full_name='xgt.VertexFrame.nvertices', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='schema', full_name='xgt.VertexFrame.schema', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error_name', full_name='xgt.VertexFrame.error_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=257,
  serialized_end=352,
)


_EDGEFRAME = _descriptor.Descriptor(
  name='EdgeFrame',
  full_name='xgt.EdgeFrame',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='xgt.EdgeFrame.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='nedges', full_name='xgt.EdgeFrame.nedges', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='schema', full_name='xgt.EdgeFrame.schema', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='source_vertex', full_name='xgt.EdgeFrame.source_vertex', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='target_vertex', full_name='xgt.EdgeFrame.target_vertex', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='source_key', full_name='xgt.EdgeFrame.source_key', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='target_key', full_name='xgt.EdgeFrame.target_key', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error_name', full_name='xgt.EdgeFrame.error_name', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=355,
  serialized_end=531,
)

_PROPERTY.fields_by_name['data_type'].enum_type = _UVALTYPEENUM
_PROPERTY.fields_by_name['role'].enum_type = _ROLEENUM
_SCHEMA.fields_by_name['property'].message_type = _PROPERTY
_TABLEFRAME.fields_by_name['schema'].message_type = _SCHEMA
_VERTEXFRAME.fields_by_name['schema'].message_type = _SCHEMA
_EDGEFRAME.fields_by_name['schema'].message_type = _SCHEMA
DESCRIPTOR.message_types_by_name['Property'] = _PROPERTY
DESCRIPTOR.message_types_by_name['Schema'] = _SCHEMA
DESCRIPTOR.message_types_by_name['TableFrame'] = _TABLEFRAME
DESCRIPTOR.message_types_by_name['VertexFrame'] = _VERTEXFRAME
DESCRIPTOR.message_types_by_name['EdgeFrame'] = _EDGEFRAME
DESCRIPTOR.enum_types_by_name['UvalTypeEnum'] = _UVALTYPEENUM
DESCRIPTOR.enum_types_by_name['RoleEnum'] = _ROLEENUM
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Property = _reflection.GeneratedProtocolMessageType('Property', (_message.Message,), dict(
  DESCRIPTOR = _PROPERTY,
  __module__ = 'SchemaMessages_pb2'
  # @@protoc_insertion_point(class_scope:xgt.Property)
  ))
_sym_db.RegisterMessage(Property)

Schema = _reflection.GeneratedProtocolMessageType('Schema', (_message.Message,), dict(
  DESCRIPTOR = _SCHEMA,
  __module__ = 'SchemaMessages_pb2'
  # @@protoc_insertion_point(class_scope:xgt.Schema)
  ))
_sym_db.RegisterMessage(Schema)

TableFrame = _reflection.GeneratedProtocolMessageType('TableFrame', (_message.Message,), dict(
  DESCRIPTOR = _TABLEFRAME,
  __module__ = 'SchemaMessages_pb2'
  # @@protoc_insertion_point(class_scope:xgt.TableFrame)
  ))
_sym_db.RegisterMessage(TableFrame)

VertexFrame = _reflection.GeneratedProtocolMessageType('VertexFrame', (_message.Message,), dict(
  DESCRIPTOR = _VERTEXFRAME,
  __module__ = 'SchemaMessages_pb2'
  # @@protoc_insertion_point(class_scope:xgt.VertexFrame)
  ))
_sym_db.RegisterMessage(VertexFrame)

EdgeFrame = _reflection.GeneratedProtocolMessageType('EdgeFrame', (_message.Message,), dict(
  DESCRIPTOR = _EDGEFRAME,
  __module__ = 'SchemaMessages_pb2'
  # @@protoc_insertion_point(class_scope:xgt.EdgeFrame)
  ))
_sym_db.RegisterMessage(EdgeFrame)


# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import AdminService_pb2 as AdminService__pb2


class AdminServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.ChangeErrorCount = channel.unary_unary(
        '/xgt.AdminService/ChangeErrorCount',
        request_serializer=AdminService__pb2.ChangeErrorCountRequest.SerializeToString,
        response_deserializer=AdminService__pb2.ChangeErrorCountResponse.FromString,
        )
    self.Identifier = channel.unary_unary(
        '/xgt.AdminService/Identifier',
        request_serializer=AdminService__pb2.IdentifierRequest.SerializeToString,
        response_deserializer=AdminService__pb2.IdentifierResponse.FromString,
        )
    self.FreeUserMemorySize = channel.unary_unary(
        '/xgt.AdminService/FreeUserMemorySize',
        request_serializer=AdminService__pb2.FreeUserMemorySizeRequest.SerializeToString,
        response_deserializer=AdminService__pb2.FreeUserMemorySizeResponse.FromString,
        )
    self.MaxUserMemorySize = channel.unary_unary(
        '/xgt.AdminService/MaxUserMemorySize',
        request_serializer=AdminService__pb2.MaxUserMemorySizeRequest.SerializeToString,
        response_deserializer=AdminService__pb2.MaxUserMemorySizeResponse.FromString,
        )
    self.Shutdown = channel.unary_unary(
        '/xgt.AdminService/Shutdown',
        request_serializer=AdminService__pb2.ShutdownRequest.SerializeToString,
        response_deserializer=AdminService__pb2.ShutdownResponse.FromString,
        )
    self.StatReport = channel.unary_unary(
        '/xgt.AdminService/StatReport',
        request_serializer=AdminService__pb2.StatReportRequest.SerializeToString,
        response_deserializer=AdminService__pb2.StatReportResponse.FromString,
        )
    self.StatReset = channel.unary_unary(
        '/xgt.AdminService/StatReset',
        request_serializer=AdminService__pb2.StatResetRequest.SerializeToString,
        response_deserializer=AdminService__pb2.StatResetResponse.FromString,
        )
    self.Version = channel.unary_unary(
        '/xgt.AdminService/Version',
        request_serializer=AdminService__pb2.VersionRequest.SerializeToString,
        response_deserializer=AdminService__pb2.VersionResponse.FromString,
        )


class AdminServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def ChangeErrorCount(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Identifier(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def FreeUserMemorySize(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def MaxUserMemorySize(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Shutdown(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def StatReport(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def StatReset(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Version(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_AdminServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'ChangeErrorCount': grpc.unary_unary_rpc_method_handler(
          servicer.ChangeErrorCount,
          request_deserializer=AdminService__pb2.ChangeErrorCountRequest.FromString,
          response_serializer=AdminService__pb2.ChangeErrorCountResponse.SerializeToString,
      ),
      'Identifier': grpc.unary_unary_rpc_method_handler(
          servicer.Identifier,
          request_deserializer=AdminService__pb2.IdentifierRequest.FromString,
          response_serializer=AdminService__pb2.IdentifierResponse.SerializeToString,
      ),
      'FreeUserMemorySize': grpc.unary_unary_rpc_method_handler(
          servicer.FreeUserMemorySize,
          request_deserializer=AdminService__pb2.FreeUserMemorySizeRequest.FromString,
          response_serializer=AdminService__pb2.FreeUserMemorySizeResponse.SerializeToString,
      ),
      'MaxUserMemorySize': grpc.unary_unary_rpc_method_handler(
          servicer.MaxUserMemorySize,
          request_deserializer=AdminService__pb2.MaxUserMemorySizeRequest.FromString,
          response_serializer=AdminService__pb2.MaxUserMemorySizeResponse.SerializeToString,
      ),
      'Shutdown': grpc.unary_unary_rpc_method_handler(
          servicer.Shutdown,
          request_deserializer=AdminService__pb2.ShutdownRequest.FromString,
          response_serializer=AdminService__pb2.ShutdownResponse.SerializeToString,
      ),
      'StatReport': grpc.unary_unary_rpc_method_handler(
          servicer.StatReport,
          request_deserializer=AdminService__pb2.StatReportRequest.FromString,
          response_serializer=AdminService__pb2.StatReportResponse.SerializeToString,
      ),
      'StatReset': grpc.unary_unary_rpc_method_handler(
          servicer.StatReset,
          request_deserializer=AdminService__pb2.StatResetRequest.FromString,
          response_serializer=AdminService__pb2.StatResetResponse.SerializeToString,
      ),
      'Version': grpc.unary_unary_rpc_method_handler(
          servicer.Version,
          request_deserializer=AdminService__pb2.VersionRequest.FromString,
          response_serializer=AdminService__pb2.VersionResponse.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'xgt.AdminService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))
